import struct
from typing import Optional

from bxcommon.constants import MSG_NULL_BYTE
from bxcommon.exceptions import ChecksumError, PayloadLenError
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.utils import crypto
from bxgateway import log_messages
from bxgateway import ont_constants
from bxutils import logging

logger = logging.get_logger(__name__)


class OntMessage(AbstractMessage):
    MESSAGE_TYPE = b"ont"
    HEADER_LENGTH = ont_constants.ONT_HDR_COMMON_OFF

    def __init__(self, magic: Optional[int] = None, command: Optional[bytes] = None, payload_len: Optional[int] = None,
                 buf: Optional[bytearray] = None):
        self.buf = buf
        # pyre-fixme[6]: Expected `Union[bytearray, bytes, memoryview]` for 1st
        #  param but got `Optional[bytearray]`.
        self._memoryview = memoryview(buf)

        # pyre-fixme[6]: Expected `int` for 1st param but got `Optional[int]`.
        checksum = crypto.double_sha256(self._memoryview[ont_constants.ONT_HDR_COMMON_OFF:payload_len +
                                        ont_constants.ONT_HDR_COMMON_OFF])

        off = 0
        # pyre-fixme[6]: Expected `Union[array.array[typing.Any], bytearray,
        #  memoryview, mmap.mmap]` for 2nd param but got `Optional[bytearray]`.
        struct.pack_into("<L12sL", buf, off, magic, command, payload_len)
        off += ont_constants.ONT_HEADER_MINUS_CHECKSUM
        # pyre-fixme[16]: `Optional` has no attribute `__setitem__`.
        buf[off:off + ont_constants.ONT_MSG_CHECKSUM_LEN] = checksum[0:4]

        self._magic = magic
        self._command = command
        self._payload_len = payload_len
        self._payload = None
        self._checksum = None

    def __repr__(self):
        return "OntMessage<command: {}, length: {}>".format(self.MESSAGE_TYPE, len(self.rawbytes()))

    @classmethod
    def unpack(cls, buf):
        magic, command, payload_length = struct.unpack_from("<L12sL", buf)
        checksum = buf[ont_constants.ONT_HEADER_MINUS_CHECKSUM:cls.HEADER_LENGTH]
        return command.rstrip(MSG_NULL_BYTE), magic, checksum, payload_length

    @classmethod
    def validate_payload(cls, buf, unpacked_args):
        command, _magic, checksum, payload_length = unpacked_args
        if payload_length != len(buf) - cls.HEADER_LENGTH:
            error_message = log_messages.PAYLOAD_LENGTH_MISMATCH.text.format(payload_length, len(buf))
            logger.error(error_message)
            raise PayloadLenError(error_message)
        ref_checksum = crypto.double_sha256(buf[cls.HEADER_LENGTH:cls.HEADER_LENGTH + payload_length])[0:4]
        if checksum != ref_checksum:
            error_message = log_messages.PACKET_CHECKSUM_MISMATCH.text.format(
                bytearray(checksum), ref_checksum, repr(buf)
            )
            logger.error(error_message)
            raise ChecksumError(error_message, buf)

    @classmethod
    def initialize_class(cls, cls_type, buf, unpacked_args):
        return cls_type(buf=buf)

    def rawbytes(self) -> memoryview:
        assert self.buf is not None
        # pyre-fixme[6]: Expected `int` for 1st param but got `Optional[int]`.
        # pyre-fixme[6]: Expected `Sized` for 1st param but got `Optional[bytearray]`.
        if self.payload_len() + ont_constants.ONT_HDR_COMMON_OFF == len(self.buf):
            return self._memoryview
        else:
            return self._memoryview[0:self.payload_len() + ont_constants.ONT_HDR_COMMON_OFF]

    def magic(self) -> int:
        if self._magic is None:
            assert self.buf is not None
            # pyre-fixme[6]: Expected `Union[array.array[int], bytearray, bytes,
            #  memoryview, mmap.mmap]` for 2nd param but got `Optional[bytearray]`.
            self._magic, = struct.unpack_from("<L", self.buf)
        magic = self._magic
        assert magic is not None
        return magic

    def command(self) -> bytes:
        if self._command is None:
            assert self.buf is not None
            # pyre-fixme[6]: Expected `Union[array.array[int], bytearray, bytes,
            #  memoryview, mmap.mmap]` for 2nd param but got `Optional[bytearray]`.
            self._command = struct.unpack_from("<12s", self.buf, 4)[0].rstrip(MSG_NULL_BYTE)
        command = self._command
        assert command is not None
        return command

    def payload_len(self) -> int:
        if self._payload_len is None:
            assert self.buf is not None
            # pyre-fixme[6]: Expected `Union[array.array[int], bytearray, bytes,
            #  memoryview, mmap.mmap]` for 2nd param but got `Optional[bytearray]`.
            self._payload_len, = struct.unpack_from("<L", self.buf, 16)
        payload_len = self._payload_len
        assert payload_len is not None
        return payload_len

    def payload(self) -> bytearray:
        if self._payload is None:
            assert self.buf is not None
            # pyre-fixme[16]: `Optional` has no attribute `__getitem__`.
            self._payload = self.buf[ont_constants.ONT_HDR_COMMON_OFF:self.payload_len() +
                                     ont_constants.ONT_HDR_COMMON_OFF]
        payload = self._payload
        assert isinstance(payload, bytearray)
        return payload

    def checksum(self):
        if self._checksum is None:
            self._checksum = self.buf[ont_constants.ONT_HEADER_MINUS_CHECKSUM:ont_constants.ONT_HDR_COMMON_OFF]
        return self._checksum



import struct

from bxcommon import constants
from bxcommon.messages.abstract_internal_message import AbstractInternalMessage
from bxcommon.messages.bloxroute.abstract_bloxroute_message import (
    AbstractBloxrouteMessage,
)
from bxcommon.messages.bloxroute.abstract_broadcast_message import (
    AbstractBroadcastMessage,
)
from bxcommon.messages.bloxroute.bloxroute_message_type import (
    BloxrouteMessageType,
)
from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.messages.bloxroute.v7.tx_message_v7 import TxMessageV7
from bxcommon.messages.versioning.abstract_message_converter import (
    AbstractMessageConverter,
)


class _TxMessageConverterV7(AbstractMessageConverter):
    _MSG_TYPE_TO_OLD_MSG_CLASS_MAPPING = {
        BloxrouteMessageType.TRANSACTION: TxMessageV7
    }

    _MSG_TYPE_TO_NEW_MSG_CLASS_MAPPING = {
        BloxrouteMessageType.TRANSACTION: TxMessage
    }

    _BASE_LENGTH = (
        AbstractBroadcastMessage.HEADER_LENGTH
        + AbstractBroadcastMessage.PAYLOAD_LENGTH
        - constants.CONTROL_FLAGS_LEN
    )
    _LEFT_BREAKPOINT = (
        _BASE_LENGTH + constants.SID_LEN + constants.QUOTA_FLAG_LEN
    )
    _RIGHT_BREAKPOINT = (
        _BASE_LENGTH
        + constants.SID_LEN
        + constants.TRANSACTION_FLAG_LEN
        + constants.UL_INT_SIZE_IN_BYTES
    )

    def convert_to_older_version(
        self, msg: AbstractInternalMessage
    ) -> AbstractInternalMessage:
        msg_type = msg.MESSAGE_TYPE

        if msg_type not in self._MSG_TYPE_TO_OLD_MSG_CLASS_MAPPING:
            raise ValueError(
                f"Tried to convert unexpected new "
                f"message type to v7: {msg_type}"
            )

        old_version_msg_class = self._MSG_TYPE_TO_OLD_MSG_CLASS_MAPPING[
            msg_type
        ]
        old_version_payload_len = (
            msg.payload_len()
            - constants.QUOTA_FLAG_LEN
            - constants.UL_INT_SIZE_IN_BYTES
        )

        old_version_msg_bytes = bytearray(
            self._LEFT_BREAKPOINT
            + len(msg.rawbytes()[self._RIGHT_BREAKPOINT :])
        )
        old_version_msg_bytes[: self._LEFT_BREAKPOINT] = msg.rawbytes()[
            : self._LEFT_BREAKPOINT
        ]
        old_version_msg_bytes[self._LEFT_BREAKPOINT :] = msg.rawbytes()[
            self._RIGHT_BREAKPOINT :
        ]

        return AbstractBloxrouteMessage.initialize_class(
            old_version_msg_class,
            old_version_msg_bytes,
            (msg_type, old_version_payload_len),
        )

    def convert_from_older_version(
        self, msg: AbstractInternalMessage
    ) -> AbstractInternalMessage:
        msg_type = msg.MESSAGE_TYPE

        if msg_type not in self._MSG_TYPE_TO_NEW_MSG_CLASS_MAPPING:
            raise ValueError(
                f"Tried to convert unexpected old message type to "
                f"v7: {msg_type}"
            )

        new_msg_class = self._MSG_TYPE_TO_NEW_MSG_CLASS_MAPPING[msg_type]
        new_payload_len = (
            msg.payload_len() +
            constants.QUOTA_FLAG_LEN +
            constants.UL_INT_SIZE_IN_BYTES
        )

        new_msg_bytes = bytearray(
            AbstractBloxrouteMessage.HEADER_LENGTH + new_payload_len
        )
        new_msg_bytes[: self._LEFT_BREAKPOINT] = msg.rawbytes()[
            : self._LEFT_BREAKPOINT
        ]

        struct.pack_into("<B", new_msg_bytes, self._LEFT_BREAKPOINT, 0)
        struct.pack_into("<L", new_msg_bytes, self._LEFT_BREAKPOINT, 0)
        new_msg_bytes[self._RIGHT_BREAKPOINT :] = msg.rawbytes()[
            self._LEFT_BREAKPOINT :
        ]

        return AbstractBloxrouteMessage.initialize_class(
            new_msg_class, new_msg_bytes, (msg_type, new_payload_len)
        )

    def convert_first_bytes_to_older_version(
        self, first_msg_bytes: memoryview
    ) -> memoryview:
        raise NotImplementedError

    def convert_first_bytes_from_older_version(
        self, first_msg_bytes: memoryview
    ) -> memoryview:
        raise NotImplementedError

    def convert_last_bytes_to_older_version(
        self, last_msg_bytes: memoryview
    ) -> memoryview:
        raise NotImplementedError

    def convert_last_bytes_from_older_version(
        self, last_msg_bytes: memoryview
    ) -> memoryview:
        raise NotImplementedError

    def get_message_size_change_to_older_version(self) -> int:
        raise NotImplementedError

    def get_message_size_change_from_older_version(self) -> int:
        raise NotImplementedError


tx_message_converter_v7 = _TxMessageConverterV7()



from bxcommon.utils.blockchain_utils.btc import btc_common_constants

BTC_MAGIC_NUMBERS = {
    "main": 0xD9B4BEF9,
    "testnet": 0xDAB5BFFA,
    "testnet3": 0x0709110B,
    "regtest": 0xDAB5BFFA,
    "namecoin": 0xFEB4BEF9
}

# The length of everything in the header minus the checksum
BTC_HEADER_MINUS_CHECKSUM = 20
BTC_HDR_COMMON_OFF = 24  # type: int
BTC_BLOCK_HDR_SIZE = 80
BTC_SHORT_NONCE_SIZE = 8
# Length of a sha256 hash
BTC_SHA_HASH_LEN = btc_common_constants.BTC_SHA_HASH_LEN
BTC_IP_ADDR_PORT_SIZE = 18
BTC_COMPACT_BLOCK_SHORT_ID_LEN = 6
BTC_VARINT_MIN_SIZE = 3

# The services that we provide
# 1: can ask for full blocks.
# 0x20: Node that is compatible with the hard fork.
BTC_CASH_SERVICE_BIT = 0x20  # Bitcoin cash service bit
BTC_NODE_SERVICES = 1
BTC_CASH_SERVICES = 33

BTC_OBJTYPE_TX = 1
BTC_OBJTYPE_BLOCK = 2
BTC_OBJTYPE_FILTERED_BLOCK = 3

BTC_HELLO_MESSAGES = [b"version", b"verack"]

# Indicator byte compressing bitcoin blocks to indicate short id
BTC_SHORT_ID_INDICATOR = 0xFF
BTC_SHORT_ID_INDICATOR_AS_BYTEARRAY = bytearray([BTC_SHORT_ID_INDICATOR])
BTC_SHORT_ID_INDICATOR_LENGTH = 1

TX_VERSION_LEN = btc_common_constants.TX_VERSION_LEN
TX_SEGWIT_FLAG_LEN = btc_common_constants.TX_SEGWIT_FLAG_LEN
TX_LOCK_TIME_LEN = btc_common_constants.TX_LOCK_TIME_LEN
TX_SEGWIT_FLAG_VALUE = btc_common_constants.TX_SEGWIT_FLAG_VALUE

NODE_WITNESS_SERVICE_FLAG = (1 << 3)

BTC_VARINT_SHORT_INDICATOR = btc_common_constants.BTC_VARINT_SHORT_INDICATOR
BTC_VARINT_SHORT_INDICATOR_AS_BYTEARRAY = bytearray([BTC_VARINT_SHORT_INDICATOR])
BTC_VARINT_INT_INDICATOR = btc_common_constants.BTC_VARINT_INT_INDICATOR
BTC_VARINT_INT_INDICATOR_AS_BYTEARRAY = bytearray([BTC_VARINT_INT_INDICATOR])
BTC_VARINT_LONG_INDICATOR = btc_common_constants.BTC_VARINT_LONG_INDICATOR
BTC_VARINT_LONG_INDICATOR_AS_BYTEARRAY = bytearray([BTC_VARINT_LONG_INDICATOR])

BTC_COMPACT_BLOCK_RECOVERY_TIMEOUT_S = 10
BTC_COMPACT_BLOCK_DECOMPRESS_MIN_TX_COUNT = 10000

BTC_DEFAULT_BLOCK_SIZE = 621000
BTC_MINIMAL_SUB_TASK_TX_COUNT = 2500



from typing import List

import blxr_rlp as rlp

from bxcommon.messages.eth.serializers.transaction import Transaction
from bxgateway.messages.eth.protocol.eth_protocol_message import EthProtocolMessage
from bxgateway.messages.eth.protocol.eth_protocol_message_type import EthProtocolMessageType


class PooledTransactionsEthProtocolMessage(EthProtocolMessage):
    msg_type = EthProtocolMessageType.POOLED_TRANSACTIONS

    fields = [
        ("transactions", rlp.sedes.CountableList(Transaction))
    ]

    transactions: List[Transaction]

    def __init__(self, msg_bytes, *args, **kwargs) -> None:
        self.transactions = []
        super().__init__(msg_bytes, *args, **kwargs)

    def get_transactions(self) -> List[Transaction]:
        return self.get_field_value("transactions")



from collections import deque
from typing import Optional, Set, Deque, Callable, TypeVar, Generic

from bxcommon.utils.memory_utils import SpecialMemoryProperties, SpecialTuple

import task_pool_executor as tpe


T = TypeVar("T", bound=tpe.MainTaskBase)


class TaskQueueProxy(Generic[T], SpecialMemoryProperties):
    QUEUE_GROW_SIZE = 1  # TODO: increase this after moving to async task execution

    def __init__(
        self,
        task_creator: Callable[[], T],
        grow_size: int = QUEUE_GROW_SIZE
    ):
        self._task_creator = task_creator
        self._grow_size = grow_size
        self._queue: Deque[T] = deque()
        self._extend_queue()

    def __len__(self) -> int:
        return len(self._queue)

    def borrow_task(self) -> T:
        try:
            tsk = self._queue.popleft()
        except IndexError:
            tsk = self._extend_queue()
        return tsk

    def return_task(self, tsk: T):
        self._queue.appendleft(tsk)

    def special_memory_size(self, ids: Optional[Set[int]] = None) -> SpecialTuple:
        total_size = 0
        if ids is None:
            ids = set()
        for tsk in self._queue:
            total_size += tsk.get_task_byte_size()
        return SpecialTuple(total_size, ids)

    def _extend_queue(self) -> T:
        tsk = None
        for _ in range(self._grow_size):
            tsk = self._task_creator()
            self._queue.append(tsk)

        assert tsk is not None
        return tsk



import struct
from hashlib import sha256
from typing import Union

from bxcommon.utils.blockchain_utils.btc import btc_common_constants
from bxcommon.utils.blockchain_utils.btc.btc_object_hash import BtcObjectHash


def btc_varint_to_int(buf, off):
    """
    Converts a varint to a regular integer in a buffer bytearray.
    https://en.bitcoin.it/wiki/Protocol_documentation#Variable_length_integer
    """
    if not isinstance(buf, memoryview):
        buf = memoryview(buf)

    if buf[off] == btc_common_constants.BTC_VARINT_LONG_INDICATOR:
        return struct.unpack_from("<Q", buf, off + 1)[0], 9
    elif buf[off] == btc_common_constants.BTC_VARINT_INT_INDICATOR:
        return struct.unpack_from("<I", buf, off + 1)[0], 5
    elif buf[off] == btc_common_constants.BTC_VARINT_SHORT_INDICATOR:
        return struct.unpack_from("<H", buf, off + 1)[0], 3
    else:
        return struct.unpack_from("B", buf, off)[0], 1


def is_segwit(buf: Union[memoryview, bytearray], off: int = 0) -> bool:
    """
    Determines if a transaction is a segwit transaction by reading the marker and flag bytes
    :param buf: the bytes of the transaction contents
    :param off: the desired offset
    :return: boolean indicating segwit
    """
    segwit_flag, = struct.unpack_from(">h", buf, off + btc_common_constants.TX_VERSION_LEN)
    return segwit_flag == btc_common_constants.TX_SEGWIT_FLAG_VALUE


def get_txid(buffer: Union[memoryview, bytearray]) -> BtcObjectHash:
    """
    Actually gets the txid, which is the same as the hash for non segwit transactions
    :param buffer: the bytes of the transaction contents
    :return: hash object
    """

    flag_len = btc_common_constants.TX_SEGWIT_FLAG_LEN if is_segwit(buffer) else 0
    txid = sha256(buffer[:btc_common_constants.TX_VERSION_LEN])
    end = btc_common_constants.TX_VERSION_LEN + flag_len
    io_size, _, _ = get_tx_io_count_and_size(buffer, end, tail=-1)
    txid.update(buffer[end:end + io_size])
    txid.update(buffer[-btc_common_constants.TX_LOCK_TIME_LEN:])

    return BtcObjectHash(buf=sha256(txid.digest()).digest(), length=btc_common_constants.BTC_SHA_HASH_LEN)


def get_tx_io_count_and_size(buf: Union[memoryview, bytearray], start, tail):
    end = start

    txin_c, size = btc_varint_to_int(buf, end)
    end += size

    if end > tail > 0:
        return -1

    for _ in range(txin_c):
        end += 36
        script_len, size = btc_varint_to_int(buf, end)
        end += size + script_len + 4

        if end > tail > 0:
            return -1

    txout_c, size = btc_varint_to_int(buf, end)
    end += size
    for _ in range(txout_c):
        end += 8
        script_len, size = btc_varint_to_int(buf, end)
        end += size + script_len

        if end > tail > 0:
            return -1

    return end - start, txin_c, txout_c



from bxcommon.connections.abstract_connection import AbstractConnection
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.services.broadcast_service import BroadcastService, BroadcastOptions


class GatewayBroadcastService(
    BroadcastService[AbstractMessage, AbstractConnection, BroadcastOptions]
):
    def should_broadcast_to_connection(
        self,
        message: AbstractMessage,
        connection: AbstractConnection,
        options: BroadcastOptions
    ) -> bool:
        # gateway does not really care about network numbers
        return True



import base64
import json
import struct
from dataclasses import dataclass
from typing import List, Optional, Union

from bxcommon.messages.abstract_block_message import AbstractBlockMessage
from bxcommon.utils import crypto, model_loader
from bxcommon.utils.blockchain_utils.ont.ont_common_utils import ont_varint_to_int, get_txid
from bxcommon.utils.blockchain_utils.ont.ont_object_hash import OntObjectHash
from bxgateway import ont_constants
from bxgateway.messages.ont.ont_message import OntMessage
from bxgateway.messages.ont.ont_message_type import OntMessageType


@dataclass
class ConsensusMsgPayload:
    type: int
    len: int
    payload: str


class OntConsensusMessage(OntMessage, AbstractBlockMessage):
    MESSAGE_TYPE = OntMessageType.CONSENSUS

    def __init__(self, magic: Optional[int] = None, version: Optional[int] = None,
                 consensus_payload: Optional[bytes] = None, buf: Optional[Union[bytearray, memoryview]] = None):

        if buf is None:
            # pyre-fixme[6]: Expected `typing.Sized` for 1st positional only parameter to call `len`
            #  but got `Optional[bytes]`.
            buf = bytearray(ont_constants.ONT_HDR_COMMON_OFF + ont_constants.ONT_INT_LEN + len(consensus_payload))
            self.buf = buf

            off = ont_constants.ONT_HDR_COMMON_OFF
            struct.pack_into("<I", buf, off, version)
            off += ont_constants.ONT_INT_LEN

            # pyre-fixme[6]: Expected `typing.Sized` for 1st positional only parameter to call `len`
            #  but got `Optional[bytes]`.
            # pyre-fixme[6]: Expected `Union[typing.Iterable[int], bytes]` for 2nd positional only parameter
            #  to call `bytearray.__setitem__` but got `Optional[bytes]`.
            buf[off:off + len(consensus_payload)] = consensus_payload
            # pyre-fixme[6]: Expected `typing.Sized` for 1st positional only parameter to call `len` but
            #  got `Optional[bytes]`.
            off += len(consensus_payload)

            super().__init__(magic, self.MESSAGE_TYPE, off - ont_constants.ONT_HDR_COMMON_OFF, buf)

        else:
            self.buf = buf
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._memoryview = memoryview(buf)
        self._version = self._consensus_payload = self._consensus_payload_header = self._owner_and_signature = None
        self._consensus_data = self._consensus_data_type = self._consensus_data_len = None
        self._consensus_data_full_len = self._consensus_data_str = self._consensus_data_json = None
        self._block_start_txns = self._block_start_len_memoryview = self._empty_block_offset = None
        self._prev_block = self._decoded_payload = self._payload_tail = self._message_tail = None
        self._tx_offset = self._txn_header = self._txn_count = self._hash_val = self._header_offset = None
        self._parsed = False

    def __repr__(self):
        return "ConsensusOntMessage<type: {}, length: {}, block hash: {}>".format(self.consensus_data_type(),
                                                                                  len(self.rawbytes()),
                                                                                  self._hash_val)

    def parse_message(self):
        if not self._parsed:
            self._parsed = True
        off = ont_constants.ONT_HDR_COMMON_OFF
        self._version, = struct.unpack_from("<I", self.buf, off)
        off += ont_constants.ONT_INT_LEN
        self._consensus_payload = self._memoryview[off:]
        off += ont_constants.ONT_HASH_LEN + 2 * ont_constants.ONT_INT_LEN + ont_constants.ONT_SHORT_LEN
        self._consensus_data_full_len, size = ont_varint_to_int(self.buf, off)
        off += size
        self._consensus_payload_header = self._memoryview[:off]
        self._consensus_data = self._memoryview[off:off + self._consensus_data_full_len]
        self._consensus_data_str = bytearray(self._consensus_data).decode()
        self._consensus_data_json = json.loads(self._consensus_data_str)
        off += self._consensus_data_full_len
        self._owner_and_signature = self._memoryview[off:]
        self._consensus_data_type = self._consensus_data_json["type"]
        if self._consensus_data_type == ont_constants.BLOCK_PROPOSAL_CONSENSUS_MESSAGE_TYPE:
            encoded_payload = model_loader.load_model(ConsensusMsgPayload, self._consensus_data_json)
            self._consensus_data_len = encoded_payload.len
            self._decoded_payload = memoryview(base64.b64decode(encoded_payload.payload))
            self.parse_block_vbft_type()

    def parse_block_vbft_type(self):
        block_start_len, size = ont_varint_to_int(self._decoded_payload, 0)
        self._block_start_len_memoryview = self._decoded_payload[:size]
        off = size
        block_start = self._decoded_payload[off:off + block_start_len]
        self._block_start_txns = self.parse_block_core_type(block_start)
        self._empty_block_offset = off + block_start_len
        self._payload_tail = self._decoded_payload[self._empty_block_offset:]

    def parse_block_core_type(self, buf: Union[bytearray, memoryview]) -> List[memoryview]:
        if isinstance(buf, bytearray):
            buf = memoryview(buf)
        off = ont_constants.ONT_INT_LEN
        self._prev_block = OntObjectHash(buf, off, ont_constants.ONT_HASH_LEN)
        off += ont_constants.ONT_HASH_LEN * 3 + ont_constants.ONT_BLOCK_TIME_HEIGHT_CONS_DATA_LEN
        consensus_payload_length, size = ont_varint_to_int(buf, off)
        off += consensus_payload_length + size + ont_constants.ONT_BLOCK_NEXT_BOOKKEEPER_LEN
        self._header_offset = off
        hash_header = buf[:self._header_offset]
        self._hash_val = OntObjectHash(buf=crypto.double_sha256(hash_header), length=ont_constants.ONT_HASH_LEN)
        bookkeepers_length, size = ont_varint_to_int(buf, off)
        off += size
        for _ in range(bookkeepers_length):
            _, size = ont_varint_to_int(buf, off)
            off += size + ont_constants.ONT_BOOKKEEPER_LEN

        sig_data_length, size = ont_varint_to_int(buf, off)
        off += size
        for _ in range(sig_data_length):
            sig_length, size = ont_varint_to_int(buf, off)
            off += size + sig_length

        self._txn_count, = struct.unpack_from("<L", buf, off)
        off += ont_constants.ONT_INT_LEN
        self._tx_offset = off
        self._txn_header = buf[:off]
        txns = []
        start = self._tx_offset
        txn_count = self._txn_count
        assert isinstance(txn_count, int)
        for _ in range(txn_count):
            _, off = get_txid(buf[start:])
            sig_length, size = ont_varint_to_int(buf[start:], off)
            off += size
            for i in range(sig_length):
                invoke_length, size = ont_varint_to_int(buf[start:], off)
                off += size + invoke_length
                verify_length, size = ont_varint_to_int(buf[start:], off)
                off += size + verify_length
            txns.append(buf[start:start + off])
            start += off

        return txns

    def block_start_len_memoryview(self) -> memoryview:
        if self._block_start_len_memoryview is None:
            self.parse_message()
        block_start_len_memoryview = self._block_start_len_memoryview
        assert isinstance(block_start_len_memoryview, memoryview)
        return block_start_len_memoryview

    def consensus_payload_header(self) -> memoryview:
        if self._consensus_payload_header is None:
            self.parse_message()
        consensus_payload_header = self._consensus_payload_header
        assert isinstance(consensus_payload_header, memoryview)
        return consensus_payload_header

    def consensus_data_len(self) -> int:
        if not self._parsed:
            self.parse_message()
        consensus_data_len = self._consensus_data_len
        assert isinstance(consensus_data_len, int)
        return consensus_data_len

    def consensus_data_type(self) -> int:
        if not self._parsed:
            self.parse_message()
        consensus_data_type = self._consensus_data_type
        assert isinstance(consensus_data_type, int)
        return consensus_data_type

    def txn_count(self) -> int:
        if not self._parsed:
            self.parse_message()
        txn_count = self._txn_count
        assert isinstance(txn_count, int)
        return txn_count

    def txn_header(self) -> memoryview:
        if self._txn_header is None:
            self.parse_message()
        txn_header = self._txn_header
        assert isinstance(txn_header, memoryview)
        return txn_header

    def txns(self) -> List[memoryview]:
        if not self._parsed:
            self.parse_message()
        block_start_txns = self._block_start_txns
        assert isinstance(block_start_txns, list)
        return block_start_txns

    def payload_tail(self) -> memoryview:
        if self._payload_tail is None:
            self.parse_message()
        payload_tail = self._payload_tail
        assert isinstance(payload_tail, memoryview)
        return payload_tail

    def block_hash(self) -> OntObjectHash:
        if not self._parsed:
            self.parse_message()
        hash_val = self._hash_val
        assert isinstance(hash_val, OntObjectHash)
        return hash_val

    def prev_block_hash(self) -> OntObjectHash:
        if not self._parsed:
            self.parse_message()
        prev_block = self._prev_block
        assert isinstance(prev_block, OntObjectHash), f"{prev_block} is {type(prev_block)}"
        return prev_block

    def owner_and_signature(self) -> memoryview:
        if self._owner_and_signature is None:
            self.parse_message()
        owner_and_signature = self._owner_and_signature
        assert isinstance(owner_and_signature, memoryview)
        return owner_and_signature

    def timestamp(self) -> int:
        pass

    def extra_stats_data(self):
        return ""



import struct
import time
from datetime import datetime
from typing import Any, List, Union, Optional

import task_pool_executor as tpe

from bxcommon import constants
from bxcommon.messages.bloxroute import transactions_info_serializer
from bxcommon.messages.bloxroute.tx_service_sync_txs_message import TxServiceSyncTxsMessage
from bxcommon.models.transaction_flag import TransactionFlag
from bxcommon.models.transaction_info import TransactionSearchResult, TransactionInfo
from bxcommon.models.transaction_key import TransactionKey, TransactionCacheKeyType
from bxcommon.services.transaction_service import TransactionService, TxSyncMsgProcessingItem
from bxcommon.services.transaction_service import TxRemovalReason
from bxcommon.utils import memory_utils, crypto
from bxcommon.utils.deprecated import deprecated
from bxcommon.utils.memory_utils import ObjectSize, SizeType
from bxcommon.utils.object_encoder import ObjectEncoder
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.proxy import task_pool_proxy
from bxcommon.utils.proxy.default_map_proxy import DefaultMapProxy
from bxcommon.utils.proxy.map_proxy import MapProxy
from bxcommon.utils.stats import hooks
from bxcommon.utils.stats.transaction_stat_event_type import TransactionStatEventType
from bxcommon.utils.stats.transaction_statistics_service import tx_stats
from bxutils import logging
from bxutils.logging import log_config
from bxutils.logging.log_record_type import LogRecordType

logger = logging.get_logger(__name__)
logger_memory_cleanup = logging.get_logger(LogRecordType.BlockCleanup, __name__)

UNKNOWN_TRANSACTION_HASH: Sha256Hash = Sha256Hash(bytearray(b"\xff" * crypto.SHA256_HASH_LEN))


class ExtensionTransactionService(TransactionService):

    def __init__(self, node, network_num) -> None:
        super(ExtensionTransactionService, self).__init__(node, network_num)
        # Log levels need to be set again to include the loggers created in this conditionally imported class
        log_config.lazy_set_log_level(node.opts.log_level_overrides)

        self.proxy = tpe.TransactionService(
            task_pool_proxy.get_pool_size(),
            self.node.opts.blockchain_networks[network_num].protocol.lower(),
            node.opts.tx_mem_pool_bucket_size,
            self._get_final_tx_confirmations_count()
        )
        raw_encoder = ObjectEncoder.raw_encoder()
        self._tx_cache_key_to_short_ids = DefaultMapProxy(
            self.proxy.tx_hash_to_short_ids(), raw_encoder, raw_encoder
        )
        self._short_id_to_tx_cache_key = MapProxy(
            self.proxy.short_id_to_tx_hash(), raw_encoder, raw_encoder
        )
        # pyre-fixme[6]: Incompatible parameter type
        content_encoder = ObjectEncoder(memoryview, tpe.InputBytes)
        self._tx_cache_key_to_contents = MapProxy(
            # pyre-fixme[6]: Incompatible parameter type
            self.proxy.tx_hash_to_contents(), raw_encoder, content_encoder
        )
        self._tx_not_seen_in_blocks = self.proxy.tx_not_seen_in_blocks()

        self._tx_hash_to_time_removed = MapProxy(
            self.proxy.tx_hash_to_time_removed(), raw_encoder, raw_encoder
        )
        self._short_id_to_time_removed = MapProxy(
            self.proxy.short_id_to_time_removed(), raw_encoder, raw_encoder
        )

    def track_seen_short_ids(self, block_hash, short_ids: List[int]) -> None:
        start_datetime = datetime.now()
        super(ExtensionTransactionService, self).track_seen_short_ids(block_hash, short_ids)
        wrapped_block_hash = tpe.Sha256(tpe.InputBytes(self._wrap_sha256(block_hash).binary))
        proxy_start_datetime = datetime.now()
        # TODO when refactoring add `block_hash` to proxy.track_seen_short_ids as
        # first parameter and change ds type in cpp
        result = self.proxy.track_seen_short_ids(wrapped_block_hash, tpe.UIntList(short_ids))
        removed_contents_size, dup_sids = result
        self.update_removed_transactions(removed_contents_size, dup_sids)
        logger_memory_cleanup.statistics(
            {
                "type": "MemoryCleanup",
                "event": "ExtensionTransactionServiceTrackSeenSummary",
                "seen_short_ids_count": len(short_ids),
                "total_content_size_removed": removed_contents_size,
                "total_duplicate_short_ids": len(dup_sids),
                "proxy_call_datetime": proxy_start_datetime,
                "data": self.get_cache_state_json(),
                "start_datetime": start_datetime,
                "block_hash": repr(block_hash)
            }
        )

    def set_final_tx_confirmations_count(self, val: int):
        super(ExtensionTransactionService, self).set_final_tx_confirmations_count(val)
        self.proxy.set_final_tx_confirmations_count(val)

    def on_block_cleaned_up(self, block_hash: Sha256Hash) -> None:
        super(ExtensionTransactionService, self).on_block_cleaned_up(block_hash)
        wrapped_block_hash = tpe.Sha256(tpe.InputBytes(block_hash.binary))
        self.proxy.on_block_cleaned_up(wrapped_block_hash)

    def get_tx_service_sync_buffer(self, sync_tx_content: bool) -> memoryview:
        byte_array_obj = self.proxy.get_tx_sync_buffer(
            self._total_tx_contents_size, sync_tx_content
        )
        return memoryview(byte_array_obj)

    def update_removed_transactions(self, removed_content_size: int, short_ids: List[int]) -> None:
        self._total_tx_contents_size -= removed_content_size
        for short_id in short_ids:
            tx_stats.add_tx_by_hash_event(
                UNKNOWN_TRANSACTION_HASH, TransactionStatEventType.TX_REMOVED_FROM_MEMORY,
                self.network_num, short_id, reason=TxRemovalReason.EXTENSION_BLOCK_CLEANUP.value
            )
            self._tx_assignment_expire_queue.remove(short_id)
            if self.node.opts.dump_removed_short_ids:
                self._removed_short_ids.add(short_id)

    @deprecated
    def assign_short_id(
        self,
        transaction_hash: Sha256Hash,
        short_id: int,
        transaction_cache_key: Optional[TransactionCacheKeyType] = None
    ) -> None:
        return self.assign_short_id_by_key(
            self.get_transaction_key(transaction_hash, transaction_cache_key),
            short_id,
        )

    def assign_short_id_by_key(
        self,
        transaction_key: TransactionKey,
        short_id: int,
    ) -> None:
        """
        Adds short id mapping for transaction and schedules an alarm to cleanup entry on expiration.
        :param transaction_key: transaction key object
        :param short_id: short id to be mapped to transaction
        """
        logger.trace("Assigning sid {} to transaction {}", short_id, transaction_key.transaction_hash)

        # pyre-fixme[6]: Expected `tpe.Sha256` got `TransactionCacheKeyType`.
        has_contents = self.proxy.assign_short_id(transaction_key.transaction_cache_key, short_id)
        self.assign_short_id_base_by_key(transaction_key, short_id, has_contents, False)

    @deprecated
    def set_transaction_contents(
        self, transaction_hash: Sha256Hash, transaction_contents: Union[bytearray, memoryview],
        transaction_cache_key: Optional[TransactionCacheKeyType] = None
    ):
        return self.set_transaction_contents_by_key(
            self.get_transaction_key(transaction_hash, transaction_cache_key),
            transaction_contents
        )

    def set_transaction_contents_by_key(
        self, transaction_key: TransactionKey, transaction_contents: Union[bytearray, memoryview]
    ):
        """
        Adds transaction contents to transaction service cache with lookup key by transaction hash

        :param transaction_key: transaction key object
        :param transaction_contents: transaction contents bytes
        """
        has_short_id, previous_size = self.proxy.set_transaction_contents(
            # pyre-fixme[6]: Expected `tpe.Sha256` got `TransactionCacheKeyType`.
            transaction_key.transaction_cache_key,
            tpe.InputBytes(transaction_contents))

        self.set_transaction_contents_base_by_key(
            transaction_key,
            has_short_id,
            previous_size,
            False,
            transaction_contents,
            len(transaction_contents)
        )

    def get_transactions(
        self,
        serialized_short_ids: Optional[bytearray] = None
    ) -> TransactionSearchResult:
        """
        Fetches all transaction info for a set of short ids.
        Short ids without a transaction entry will be omitted.
        Function allows to pass a single short id or serialized list of short ids
        :param serialized_short_ids: instance of get transactions message
        :return: list of found and missing short ids
        """

        assert serialized_short_ids is not None
        input_bytes = tpe.InputBytes(serialized_short_ids)

        result_bytes = self.proxy.get_transactions_by_short_ids(input_bytes)
        assert result_bytes is not None
        result_memory_view = memoryview(result_bytes)

        found_txs_size, = struct.unpack_from("<L", result_memory_view, 0)

        txs_bytes = result_memory_view[constants.UL_INT_SIZE_IN_BYTES:constants.UL_INT_SIZE_IN_BYTES + found_txs_size]
        found_txs_info = transactions_info_serializer \
            .deserialize_transactions_info(txs_bytes)

        missing_txs_info = []
        offset = constants.UL_INT_SIZE_IN_BYTES + found_txs_size

        missing_txs_count, = struct.unpack_from("<L", result_memory_view, offset)
        offset += constants.UL_INT_SIZE_IN_BYTES

        for _ in range(missing_txs_count):
            tx_sid, = struct.unpack_from("<L", result_memory_view, offset)
            offset += constants.UL_INT_SIZE_IN_BYTES

            has_hash, = struct.unpack_from("<B", result_memory_view, offset)
            offset += constants.UL_TINY_SIZE_IN_BYTES

            if has_hash:
                tx_hash = Sha256Hash(result_memory_view[offset:offset + crypto.SHA256_HASH_LEN])
                offset += crypto.SHA256_HASH_LEN
            else:
                tx_hash = None

            missing_txs_info.append(TransactionInfo(tx_hash, None, tx_sid))

        return TransactionSearchResult(found_txs_info, missing_txs_info)

    def process_tx_sync_message(self, msg: TxServiceSyncTxsMessage) -> List[TxSyncMsgProcessingItem]:
        input_bytes = tpe.InputBytes(msg.rawbytes())
        result_bytes = self.proxy.process_tx_sync_message(input_bytes)

        assert result_bytes is not None
        result_memory_view = memoryview(result_bytes)

        result_items = []
        offset = 0

        txs_count, = struct.unpack_from("<L", result_memory_view, offset)
        offset += constants.UL_INT_SIZE_IN_BYTES

        for _ in range(txs_count):
            transaction_hash = Sha256Hash(result_memory_view[offset:offset + crypto.SHA256_HASH_LEN])
            transaction_key = self.get_transaction_key(transaction_hash)
            offset += crypto.SHA256_HASH_LEN

            content_len, = struct.unpack_from("<L", result_memory_view, offset)
            offset += constants.UL_INT_SIZE_IN_BYTES

            short_id_count, = struct.unpack_from("<L", result_memory_view, offset)
            offset += constants.UL_INT_SIZE_IN_BYTES

            if content_len > 0:
                self.set_transaction_contents_base_by_key(
                    transaction_key,
                    short_id_count > 0,
                    0,
                    False,
                    None,
                    content_len
                )

            short_ids = []

            for _ in range(short_id_count):
                short_id, = struct.unpack_from("<L", result_memory_view, offset)
                offset += constants.UL_INT_SIZE_IN_BYTES

                self.assign_short_id_base_by_key(
                    transaction_key,
                    short_id,
                    content_len > 0,
                    False
                )

                short_ids.append(short_id)

            transaction_flags = []

            for _ in range(short_id_count):
                transaction_flag, = struct.unpack_from("<H", result_memory_view, offset)
                offset += constants.TRANSACTION_FLAG_LEN
                transaction_flags.append(TransactionFlag(transaction_flag))

            result_items.append(TxSyncMsgProcessingItem(transaction_hash, content_len, short_ids, transaction_flags))

        return result_items

    def log_tx_service_mem_stats(self, include_data_structure_memory: bool = False) -> None:
        super(ExtensionTransactionService, self).log_tx_service_mem_stats(include_data_structure_memory)

        if include_data_structure_memory:
            hooks.add_obj_mem_stats(
                self.__class__.__name__,
                self.network_num,
                self._tx_not_seen_in_blocks,
                "tx_not_seen_in_blocks",
                self.get_collection_mem_stats(SizeType.OBJECT, self._tx_not_seen_in_blocks),
                object_item_count=len(self._tx_not_seen_in_blocks),
                object_type=memory_utils.ObjectType.BASE,
                size_type=SizeType.OBJECT
            )

    def get_collection_mem_stats(self, size_type: SizeType, collection_obj: Any, estimated_size: int = 0) -> ObjectSize:
        if self.get_object_type(collection_obj) == memory_utils.ObjectType.DEFAULT_MAP_PROXY:
            collection_size = collection_obj.map_obj.get_bytes_length()
            if collection_obj is self._tx_cache_key_to_short_ids:
                collection_size += (
                    len(self._short_id_to_tx_cache_key) * constants.UL_INT_SIZE_IN_BYTES)
            return memory_utils.ObjectSize(size=collection_size, flat_size=0, is_actual_size=True)
        else:
            return super(ExtensionTransactionService, self).get_collection_mem_stats(
                size_type, collection_obj, estimated_size
            )

    def get_object_type(self, collection_obj: Any):
        super(ExtensionTransactionService, self).get_object_type(collection_obj)
        if isinstance(collection_obj, DefaultMapProxy):
            return memory_utils.ObjectType.DEFAULT_MAP_PROXY
        elif isinstance(collection_obj, MapProxy):
            return memory_utils.ObjectType.MAP_PROXY
        else:
            return memory_utils.ObjectType.BASE

    def get_oldest_removed_tx_hash(self, tx_hashes: List[Sha256Hash]) -> float:
        oldest_removed_tx_hash = 0
        for tx_hash in tx_hashes:
            if tx_hash in self._tx_hash_to_time_removed and \
                    oldest_removed_tx_hash < self._tx_hash_to_time_removed[tx_hash]:
                oldest_removed_tx_hash = self._tx_hash_to_time_removed[tx_hash]
        return oldest_removed_tx_hash

    def get_oldest_removed_short_id(self, short_ids: List[int]) -> float:
        oldest_removed_short_id = 0
        for short_id in short_ids:
            if short_id in self._short_id_to_time_removed and \
                    oldest_removed_short_id < self._short_id_to_time_removed[short_id]:
                oldest_removed_short_id = self._short_id_to_time_removed[short_id]
        return oldest_removed_short_id

    def _tx_hash_to_cache_key(self, transaction_hash) -> tpe.Sha256:  # pyre-ignore
        if isinstance(transaction_hash, Sha256Hash):
            return tpe.Sha256(tpe.InputBytes(transaction_hash.binary))

        if isinstance(transaction_hash, (bytes, bytearray, memoryview)):
            return tpe.Sha256(tpe.InputBytes(transaction_hash))

        if isinstance(transaction_hash, tpe.Sha256):
            return transaction_hash

        raise ValueError("Attempted to find cache entry with incorrect key type")

        # return transaction_hash

    def _tx_cache_key_to_hash(self, transaction_cache_key) -> Sha256Hash:
        if isinstance(transaction_cache_key, Sha256Hash):
            return transaction_cache_key

        if isinstance(transaction_cache_key, (bytes, bytearray, memoryview)):
            return Sha256Hash(transaction_cache_key)

        return Sha256Hash(bytearray(transaction_cache_key.binary()))

    def _track_seen_transaction(self, transaction_cache_key):
        super(ExtensionTransactionService, self)._track_seen_transaction(transaction_cache_key)
        self.proxy.track_seen_transaction(transaction_cache_key)

    def remove_transaction_by_short_id(self, short_id: int, remove_related_short_ids: bool = False,
                                       force: bool = False,
                                       removal_reason: TxRemovalReason = TxRemovalReason.UNKNOWN):
        # overriding this in order to handle removes triggered by either the mem limit or expiration queue
        # if the remove_related_short_ids is True than we assume the call originated by the track seen call
        # else we assume it was triggered by the cleanup.
        # this is only a temporary fix and the whole class hierarchy requires some refactoring!
        if remove_related_short_ids:
            self._tx_assignment_expire_queue.remove(short_id)
            tx_stats.add_tx_by_hash_event(
                UNKNOWN_TRANSACTION_HASH, TransactionStatEventType.TX_REMOVED_FROM_MEMORY,
                self.network_num, short_id, reason=removal_reason.value
            )
            if self.node.opts.dump_removed_short_ids:
                self._removed_short_ids.add(short_id)
        else:
            super(
                ExtensionTransactionService, self
            ).remove_transaction_by_short_id(
                short_id,
                force=force,
                removal_reason=removal_reason
            )

    def clear(self):
        self.proxy.clear()

        self._short_id_to_tx_flag.clear()
        self.tx_hashes_without_content.clear()
        self.tx_hashes_without_short_id.clear()
        self._tx_assignment_expire_queue.clear()
        self._total_tx_contents_size = 0

    def _cleanup_removed_transactions_history(self):
        logger.trace(
            "Starting to cleanup transaction cache history for network "
            "number: {}.",
            self.network_num
        )
        current_time = time.time()
        tx_hash_history_len_before = len(self._tx_hash_to_time_removed)
        self._tx_hash_to_time_removed.map_obj.cleanup_removed_hashes_history(
            current_time, self._removed_txs_expiration_time_s, constants.REMOVED_TRANSACTIONS_HISTORY_LENGTH_LIMIT
        )
        tx_hash_history_len_after = len(self._tx_hash_to_time_removed)

        short_id_history_len_before = len(self._short_id_to_time_removed)
        self._short_id_to_time_removed.map_obj.cleanup_removed_short_ids_history(
            current_time, self._removed_txs_expiration_time_s, constants.REMOVED_TRANSACTIONS_HISTORY_LENGTH_LIMIT
        )
        short_id_history_len_after = len(self._short_id_to_time_removed)
        logger.trace(
            "Finished cleanup transaction cache history. "
            "tx_hash size before: {}, size after: {}."
            "short_id size before: {}, size after: {}.",
            tx_hash_history_len_before,
            tx_hash_history_len_after,
            short_id_history_len_before,
            short_id_history_len_after
        )

        return constants.REMOVED_TRANSACTIONS_HISTORY_CLEANUP_INTERVAL_S



from abc import abstractmethod, ABCMeta
from typing import Optional

from bxcommon.utils.buffers.input_buffer import InputBuffer


class AbstractMessageValidator:
    """
    Base class for message validators.
    Checks correctness of the message on different stages of processing
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def validate(
        self,
        is_full_msg: bool,
        msg_type: Optional[bytes],
        header_len: int,
        payload_len: Optional[int],
        input_buffer: InputBuffer
    ) -> None:
        """
        Validates message payload length.
        Throws MessageValidationError is message is not valid

        :param is_full_msg: indicates if the full message is available on input buffer
        :param msg_type: message type
        :param header_len: message header length
        :param payload_len: message payload length
        :param input_buffer: input buffer
        """



import time
from collections import deque
from typing import Deque, Optional, TYPE_CHECKING

from bxcommon.messages.abstract_block_message import AbstractBlockMessage
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.utils.stats.transaction_statistics_service import tx_stats
from bxutils import logging
from bxutils.logging.log_level import LogLevel

logger = logging.get_logger(__name__)

if TYPE_CHECKING:
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.connections.abstract_connection import AbstractConnection


class MessageTrackerEntry:
    message: Optional[AbstractMessage]
    label: Optional[str]
    sent_bytes: int = 0
    length: int
    queued_time: float
    last_operation_time: float

    def __init__(
        self,
        message: Optional[AbstractMessage],
        length: int,
        label: Optional[str],
    ):
        self.message = message
        self.length = length
        self.queued_time = time.time()
        self.last_operation_time = 0.0
        self.label = label

    def message_log_level(self) -> LogLevel:
        if self.message:
            if isinstance(self.message, TxMessage):
                if tx_stats.should_log_event_for_tx(
                    # pyre-fixme[16]: `Optional` has no attribute `tx_hash`.
                    self.message.tx_hash().binary,
                    # pyre-fixme[16]: `Optional` has no attribute `network_num`.
                    self.message.network_num(),
                    # pyre-fixme[16]: `Optional` has no attribute `short_id`.
                    self.message.short_id(),
                ):
                    return LogLevel.DEBUG

            # pyre-fixme[16]: `Optional` has no attribute `log_level`.
            return self.message.log_level()

        return LogLevel.DEBUG

    def as_str(self) -> str:
        base = "Message"

        if self.message is None and self.label is not None:
            base = self.label
        if self.message is not None and self.label is None:
            base = self.message
        if self.message is not None and self.label is not None:
            base = f"{self.message}({self.label})"

        return f"{base}(len: {self.length})"

    def __repr__(self):
        return (
            f"MessageTrackerEntry<"
            f"message: {self.message}, "
            f"sent_bytes: {self.sent_bytes}, "
            f"length: {self.length},"
            f"label: {self.label}>"
        )


class MessageTracker:
    """
    Service to track when message bytes get fully written from the output buffer to the
    OS level socket.
    """

    messages: Deque[MessageTrackerEntry]
    connection: "AbstractConnection"
    is_working: bool = True
    bytes_remaining: int = 0

    def __init__(self, connection: "AbstractConnection") -> None:
        self.connection = connection
        self.messages = deque()

    def __repr__(self):
        return (
            f"MessageTracker<connection: {self.connection}, "
            f"messages: {self.messages}>"
        )

    def is_sending_block_message(self) -> bool:
        if not self.messages:
            return False

        entry = self.messages[0]
        return entry.message is not None and isinstance(
            entry.message, AbstractBlockMessage
        )

    def advance_bytes(self, num_bytes: int):
        if not self.is_working:
            return

        bytes_left = num_bytes
        self.bytes_remaining -= num_bytes
        while bytes_left > 0:

            if not self.messages:
                self.connection.log_debug(
                    "Message tracker somehow got out of sync. "
                    "Attempted to send {} bytes when none left in tracker. "
                    "Resetting. ",
                    bytes_left,
                )
                self.bytes_remaining = 0
                self.messages.clear()
                return

            curr_time = time.time()
            if bytes_left >= (
                self.messages[0].length - self.messages[0].sent_bytes
            ):
                sent_message = self.messages.popleft()
                self.connection.log(
                    sent_message.message_log_level(),
                    "Sent {} to socket. Took {:.2f}ms. "
                    "last sent operation took {:.2f}ms. "
                    "{} bytes remaining on buffer.",
                    sent_message.as_str(),
                    1000 * (curr_time -
                            sent_message.last_operation_time
                            if sent_message.last_operation_time
                            else 0
                            ),
                    1000 * (curr_time - sent_message.queued_time),
                    self.bytes_remaining,
                )
                bytes_left -= sent_message.length - sent_message.sent_bytes
            else:
                in_progress_message = self.messages[0]
                in_progress_message.sent_bytes += bytes_left
                self.connection.log(
                    in_progress_message.message_log_level(),
                    "Sent {} out of {} bytes of {} to socket. "
                    "Last send operation took {:.2f}ms. "
                    "Elapsed time: {:.2f}ms. {} bytes remaining on buffer.",
                    in_progress_message.sent_bytes,
                    in_progress_message.length,
                    in_progress_message.as_str(),
                    1000 * (curr_time -
                            in_progress_message.last_operation_time
                            if in_progress_message.last_operation_time
                            else 0
                            ),
                    1000 * (curr_time - in_progress_message.queued_time),
                    self.bytes_remaining,
                )
                in_progress_message.last_operation_time = curr_time
                bytes_left = 0

    def append_message(
        self,
        num_bytes: int,
        message: Optional[AbstractMessage],
        label: Optional[str] = None,
    ):
        """
        Appends a message entry to the tracker.

        This method trusts that that num_bytes matches the message, but does
        not verify it. This is useful for Ethereum, which frames and encrypts
        the message, which may change the length of the message.
        """
        if not self.is_working:
            return

        self.messages.append(MessageTrackerEntry(message, num_bytes, label))
        self.bytes_remaining += num_bytes

    def prepend_message(
        self,
        num_bytes: int,
        message: Optional[AbstractMessage],
        label: Optional[str] = None,
    ):
        """
        Appends a message entry to the tracker.

        This method trusts that that num_bytes matches the message, but does
        not verify it. This is useful for Ethereum, which frames and encrypts
        the message, which may change the length of the message.
        """
        if not self.is_working:
            return

        if self.messages and self.messages[0].sent_bytes != 0:
            in_progress_message = self.messages.popleft()
            self.messages.appendleft(
                MessageTrackerEntry(message, num_bytes, label)
            )
            self.messages.appendleft(in_progress_message)
        else:
            self.messages.append(MessageTrackerEntry(message, num_bytes, label))

        self.bytes_remaining += num_bytes

    def empty_bytes(self, skip_bytes: int):
        """
        Remove bytes from tracker starting at `skip_bytes`.

        Used when the output buffer is being emptied, to stop tracking of
        later bytes.
        """
        bytes_skipped = 0
        index = 0

        while self.messages:
            entry = self.messages[index]
            bytes_left = entry.length - entry.sent_bytes
            bytes_skipped += bytes_left
            index += 1
            if bytes_skipped == skip_bytes:
                break
            if bytes_skipped > skip_bytes:
                index -= 1
                break

        curr_time = time.time()
        while len(self.messages) > index:
            entry_removed = self.messages.pop()
            self.bytes_remaining -= (
                entry_removed.length - entry_removed.sent_bytes
            )
            self.connection.log(
                entry_removed.message_log_level(),
                "Removed {} bytes of {} from buffer. "
                "Last operation was {:.2f}ms. ago. "
                "Message was queued for {:.2f}ms.",
                entry_removed.length,
                entry_removed.message,
                1000 * (curr_time -
                        entry_removed.last_operation_time
                        if entry_removed.last_operation_time
                        else 0
                        ),
                1000 * (curr_time - entry_removed.queued_time),
            )



from dataclasses import dataclass


@dataclass
class MessageSizeValidationSettings:
    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    max_block_size_bytes: int = None
    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    max_tx_size_bytes: int = None



from dataclasses import dataclass
from typing import Optional

from bxcommon import constants
from bxcommon.models.node_type import NodeType
from bxcommon.models.platform_provider import PlatformProvider
from bxutils import constants as util_constants


@dataclass
class NodeModel:
    # pyre-fixme[8]: Attribute has type `NodeType`; used as `None`.
    node_type: NodeType = None
    external_port: int = 0
    non_ssl_port: int = 0
    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    external_ip: str = None

    # Whether the node is online.
    online: Optional[bool] = False

    # Whether node has active connection with SDN
    sdn_connection_alive: bool = False

    # TODO: Remove this attribute as it's not being used anymore
    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    network: str = None

    # Internal id for distinguishing nodes.
    node_id: Optional[str] = None

    # The starting and ending Transaction Short ID range, inclusive.
    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    sid_start: int = None
    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    sid_end: int = None
    next_sid_start: Optional[int] = None
    next_sid_end: Optional[int] = None

    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    sid_expire_time: int = None
    last_pong_time: float = 0
    is_gateway_miner: bool = False
    is_internal_gateway: bool = False

    # Current build's version
    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    source_version: str = None

    # Bloxroute protocol version
    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    protocol_version: int = None

    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    blockchain_network_num: int = None

    # IP address of the blockchain node
    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    blockchain_ip: str = None

    # Port of the blockchain node
    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    blockchain_port: int = None

    # Nodes hostname
    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    hostname: str = None

    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    sdn_id: str = None

    # Nodes OS version
    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    os_version: str = None

    continent: Optional[str] = None
    # pyre-fixme[8]: Attribute has type `bool`; used as `None`.
    split_relays: bool = None
    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    country: str = None
    region: Optional[str] = None

    # idx enforces the one-way connection order of relays.
    # They connect to only other relays with an idx less than their own.
    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    idx: int = None

    has_fully_updated_tx_service: bool = False
    sync_txs_status = has_fully_updated_tx_service
    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    node_start_time: str = None

    # Ethereum remote blockchain attribute
    # Ethereum public key for remote blockchain connection
    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    node_public_key: str = None

    # number of redundant forwarding routes a particular relay expects by default
    baseline_route_redundancy: int = 0

    # number of redundant forwarding routes a particular relay expects to send to by default
    baseline_source_redundancy: int = 0

    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    private_ip: str = None
    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    csr: str = None
    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    cert: str = None
    # pyre-fixme[8]: Attribute has type `PlatformProvider`; used as `None`.
    platform_provider: PlatformProvider = None

    account_id: Optional[str] = None

    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    latest_source_version: str = None
    should_update_source_version: bool = False

    assigning_short_ids: Optional[bool] = False

    # blockchain node privileges for segmenting gateway types within a blockchain network
    # this property is ignored for relays
    node_privileges: str = util_constants.DEFAULT_NODE_PRIVILEGES

    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    first_seen_time: str = None

    is_docker: bool = False
    using_private_ip_connection: bool = False

    private_node: bool = False

    def __post_init__(self):
        self.sid_expire_time = constants.SID_EXPIRE_TIME_SECONDS
        # TODO: Remove network attribute, not being used
        if self.network is None:
            self.network = constants.DEFAULT_NETWORK_NAME
        if self.continent not in constants.DEFAULT_LIST_LOCATION_ORDER:
            self.continent = None
        if self.country:
            self.country = self.country[:constants.MAX_COUNTRY_LENGTH]

    def __eq__(self, other) -> bool:
        return isinstance(other, NodeModel) and other.node_id == self.node_id

    def __hash__(self):
        return hash(self.node_id)



from typing import Optional

from bxcommon import constants
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxgateway.messages.gateway.gateway_message_type import GatewayMessageType
from bxutils.logging import LogLevel


class RequestTxStreamMessage(AbstractBloxrouteMessage):

    MESSAGE_TYPE = GatewayMessageType.REQUEST_TX_STREAM

    def __init__(self, buf: Optional[bytearray] = None) -> None:
        if buf is None:
            buf = bytearray(
                AbstractBloxrouteMessage.HEADER_LENGTH
                + constants.CONTROL_FLAGS_LEN
            )

        super().__init__(self.MESSAGE_TYPE, constants.CONTROL_FLAGS_LEN, buf)

    def log_level(self):
        return LogLevel.DEBUG



from typing import Optional, Type, NamedTuple

from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.messages.abstract_message_factory import AbstractMessageFactory
from bxcommon.messages.bloxroute.bdn_performance_stats_message import BdnPerformanceStatsMessage
from bxcommon.messages.bloxroute.blockchain_network_message import RefreshBlockchainNetworkMessage
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.ack_message import AckMessage
from bxcommon.messages.bloxroute.block_confirmation_message import BlockConfirmationMessage
from bxcommon.messages.bloxroute.block_holding_message import BlockHoldingMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.broadcast_message import BroadcastMessage
from bxcommon.messages.bloxroute.compressed_block_txs_message import CompressedBlockTxsMessage
from bxcommon.messages.bloxroute.disconnect_relay_peer_message import DisconnectRelayPeerMessage
from bxcommon.messages.bloxroute.get_compressed_block_txs_message import GetCompressedBlockTxsMessage
from bxcommon.messages.bloxroute.get_tx_contents_message import GetTxContentsMessage
from bxcommon.messages.bloxroute.get_txs_message import GetTxsMessage
from bxcommon.messages.bloxroute.hello_message import HelloMessage
from bxcommon.messages.bloxroute.key_message import KeyMessage
from bxcommon.messages.bloxroute.notification_message import NotificationMessage
from bxcommon.messages.bloxroute.ping_message import PingMessage
from bxcommon.messages.bloxroute.pong_message import PongMessage
from bxcommon.messages.bloxroute.routing_update_message import RoutingUpdateMessage
from bxcommon.messages.bloxroute.transaction_cleanup_message import TransactionCleanupMessage
from bxcommon.messages.bloxroute.tx_contents_message import TxContentsMessage
from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.messages.bloxroute.tx_service_sync_blocks_short_ids_message import \
    TxServiceSyncBlocksShortIdsMessage
from bxcommon.messages.bloxroute.tx_service_sync_complete_message import \
    TxServiceSyncCompleteMessage
from bxcommon.messages.bloxroute.tx_service_sync_req_message import TxServiceSyncReqMessage
from bxcommon.messages.bloxroute.tx_service_sync_txs_message import TxServiceSyncTxsMessage
from bxcommon.messages.bloxroute.txs_message import TxsMessage
from bxcommon.models.broadcast_message_type import BroadcastMessageType

from bxcommon.utils.object_hash import ConcatHash, Sha256Hash


class BroadcastMessagePreview(NamedTuple):
    is_full_header: bool
    block_hash: Optional[Sha256Hash]
    broadcast_type: Optional[BroadcastMessageType]
    message_id: Optional[ConcatHash]
    network_num: Optional[int]
    source_id: Optional[str]
    payload_length: Optional[int]


class _BloxrouteMessageFactoryV22(AbstractMessageFactory):
    _MESSAGE_TYPE_MAPPING = {
        BloxrouteMessageType.HELLO: HelloMessage,
        BloxrouteMessageType.ACK: AckMessage,
        BloxrouteMessageType.PING: PingMessage,
        BloxrouteMessageType.PONG: PongMessage,
        BloxrouteMessageType.BROADCAST: BroadcastMessage,
        BloxrouteMessageType.TRANSACTION: TxMessage,
        BloxrouteMessageType.GET_TRANSACTIONS: GetTxsMessage,
        BloxrouteMessageType.TRANSACTIONS: TxsMessage,
        BloxrouteMessageType.GET_TX_CONTENTS: GetTxContentsMessage,
        BloxrouteMessageType.TX_CONTENTS: TxContentsMessage,
        BloxrouteMessageType.KEY: KeyMessage,
        BloxrouteMessageType.BLOCK_HOLDING: BlockHoldingMessage,
        BloxrouteMessageType.DISCONNECT_RELAY_PEER: DisconnectRelayPeerMessage,
        BloxrouteMessageType.TX_SERVICE_SYNC_REQ: TxServiceSyncReqMessage,
        BloxrouteMessageType.TX_SERVICE_SYNC_BLOCKS_SHORT_IDS: TxServiceSyncBlocksShortIdsMessage,
        BloxrouteMessageType.TX_SERVICE_SYNC_TXS: TxServiceSyncTxsMessage,
        BloxrouteMessageType.TX_SERVICE_SYNC_COMPLETE: TxServiceSyncCompleteMessage,
        BloxrouteMessageType.BLOCK_CONFIRMATION: BlockConfirmationMessage,
        BloxrouteMessageType.TRANSACTION_CLEANUP: TransactionCleanupMessage,
        BloxrouteMessageType.NOTIFICATION: NotificationMessage,
        BloxrouteMessageType.BDN_PERFORMANCE_STATS: BdnPerformanceStatsMessage,
        BloxrouteMessageType.REFRESH_BLOCKCHAIN_NETWORK: RefreshBlockchainNetworkMessage,
        BloxrouteMessageType.GET_COMPRESSED_BLOCK_TXS: GetCompressedBlockTxsMessage,
        BloxrouteMessageType.COMPRESSED_BLOCK_TXS: CompressedBlockTxsMessage,
        BloxrouteMessageType.ROUTING_UPDATE: RoutingUpdateMessage,
    }

    def __init__(self) -> None:
        super(_BloxrouteMessageFactoryV22, self).__init__(self._MESSAGE_TYPE_MAPPING)

    def get_base_message_type(self) -> Type[AbstractMessage]:
        return AbstractBloxrouteMessage


bloxroute_message_factory_v22 = _BloxrouteMessageFactoryV22()



from typing import TYPE_CHECKING, Optional
import base64

from aiohttp import web
from aiohttp.web import Request, Response
from aiohttp.web_exceptions import HTTPClientError
from prometheus_client import REGISTRY
from prometheus_client import exposition as prometheus_client

from bxcommon.rpc import rpc_constants
from bxcommon.rpc.abstract_ws_rpc_handler import AbstractWsRpcHandler
from bxcommon.rpc.https import abstract_http_rpc_server
from bxcommon.rpc.https.abstract_http_rpc_server import AbstractHttpRpcServer
from bxcommon.rpc.https.http_rpc_handler import HttpRpcHandler
from bxcommon.rpc.rpc_errors import RpcAccountIdError
from bxgateway.rpc.https.gateway_http_rpc_handler import GatewayHttpRpcHandler
from bxgateway.rpc.https.gateway_ws_handler import GatewayWsHandler

from bxutils import logging
from bxutils.encoding.json_encoder import Case

if TYPE_CHECKING:
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode


logger = logging.get_logger(__name__)


class GatewayHttpRpcServer(AbstractHttpRpcServer["AbstractGatewayNode"]):
    def __init__(self, node: "AbstractGatewayNode") -> None:
        super().__init__(node)
        self._app.add_routes(
            [
                web.get("/metrics", self.handle_metrics),
            ]
        )

    def authenticate_request(self, request: Request) -> None:
        is_authenticated = True
        if self._encoded_auth is not None:
            # pyre-fixme[16]: Callable `headers` has no attribute `__getitem__`.
            if rpc_constants.AUTHORIZATION_HEADER_KEY in request.headers:
                is_authenticated = (
                    self._encoded_auth == request.headers[rpc_constants.AUTHORIZATION_HEADER_KEY]
                )
            else:
                is_authenticated = False
        if not is_authenticated:
            message = "Request credentials are invalid. Specify RPC username and password"
            raise RpcAccountIdError(None, message)

    def request_handler(self) -> HttpRpcHandler:
        return GatewayHttpRpcHandler(self.node)

    def request_ws_handler(self) -> Optional[AbstractWsRpcHandler]:
        return GatewayWsHandler(self.node, self.node.feed_manager, Case.CAMEL)

    async def handle_metrics(self, request: Request) -> Response:
        """
        Endpoint for fetching metrics from Prometheus.

        Request can specify a custom encoding type `application/openmetrics-text` to
        fetch data in a different format.
        :param request:
        :return response:
        """
        try:
            self.authenticate_request(request)
            # pyre-fixme[16]: Callable `headers` has no attribute `get`.
            encoder, content_type = prometheus_client.choose_encoder(request.headers.get("Accept"))
            output = encoder(REGISTRY)
            response = Response(body=output, headers={"Content-Type": content_type})
            return response
        except HTTPClientError as e:
            return abstract_http_rpc_server.format_http_error(e, self._handler.content_type)

    def set_encoded_auth(self):
        encoded_auth = \
            base64.b64encode(f"{self.node.opts.rpc_user}:{self.node.opts.rpc_password}".encode("utf-8")).decode("utf-8")

        return encoded_auth



from typing import TYPE_CHECKING

from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxgateway.connections.btc.btc_relay_connection import BtcRelayConnection
from bxutils import logging

if TYPE_CHECKING:
    from bxgateway.connections.btc.btc_gateway_node import BtcGatewayNode

logger = logging.get_logger(__name__)


class BtcLossyRelayConnection(BtcRelayConnection):
    def __init__(self, sock: AbstractSocketConnectionProtocol, node: "BtcGatewayNode"):
        super(BtcLossyRelayConnection, self).__init__(sock, node)

        logger.debug("Test mode: Client is started in test mode. Simulating dropped transactions.")

        self.tx_drop_counter = 0
        self.tx_assign_drop_counter = 0
        self.tx_unknown_txs_drop_counter = 0

    def msg_tx(self, msg):

        self.tx_drop_counter += 1

        # Drop every 10th message
        if self.tx_drop_counter > 0 and self.tx_drop_counter % 10 == 0:
            self.tx_drop_counter = 0
            logger.debug("Test mode: Dropping transaction message.")
        else:
            super(BtcLossyRelayConnection, self).msg_tx(msg)

    def msg_txs(self, msg):
        self.tx_unknown_txs_drop_counter += 1

        # Drop every 3rd message
        if self.tx_unknown_txs_drop_counter > 0 and self.tx_unknown_txs_drop_counter % 3 == 0:
            self.tx_unknown_txs_drop_counter = 0
            logger.debug("Test mode: Dropping unknown txs message.")
        else:
            super(BtcLossyRelayConnection, self).msg_txs(msg)



import time

FACTOR = pow(10, 6)


def get_nonce() -> int:
    """
    using timestamp to generate unique nonce values
    :return: int
    """
    return int(time.time() * FACTOR)


def get_timestamp_from_nonce(nonce: int) -> float:
    return nonce / FACTOR



from collections import defaultdict
from dataclasses import dataclass
from typing import Dict

from bxcommon import constants


@dataclass
class BucketInfo:
    count: int
    max: int


class TransactionShortIdBuckets:
    buckets: Dict[int, BucketInfo]

    def __init__(self):
        self.buckets = defaultdict(lambda: BucketInfo(0, 0))

    def incr_short_id(self, short_id: int) -> None:
        bucket = short_id - (short_id % constants.TX_SID_INTERVAL)
        existing_bucket_info = self.buckets[bucket]
        if short_id > existing_bucket_info.max:
            existing_bucket_info.max = short_id
        existing_bucket_info.count += 1

    def __repr__(self):
        return repr(self.buckets)



from bxcommon.models.serializeable_enum import SerializeableEnum


class NetworkContent(SerializeableEnum):
    BLOCK_CONTENT = "NetworkContentBlock"
    BLOCK_TRANSACTIONS = "NetworkContentBlockTransactions"
    TRANSACTION_CONTENT = "NetworkContentTransaction"



from bxcommon.messages.bloxroute.abstract_cleanup_message import AbstractCleanupMessage
from bxcommon.services.transaction_service import TransactionService
from bxcommon.services.transaction_service import TxRemovalReason
from bxutils import logging
from bxutils.logging.log_record_type import LogRecordType

logger = logging.get_logger(LogRecordType.TransactionCleanup, __name__)


def contents_cleanup(
    transaction_service: TransactionService,
    block_confirmation_message: AbstractCleanupMessage
):
    message_hash = block_confirmation_message.message_hash()
    for short_id in block_confirmation_message.short_ids():
        transaction_service.remove_transaction_by_short_id(
            short_id,
            remove_related_short_ids=True,
            force=True,
            removal_reason=TxRemovalReason.BLOCK_CLEANUP
        )
    for tx_hash in block_confirmation_message.transaction_hashes():
        transaction_service.remove_transaction_by_key(transaction_service.get_transaction_key(tx_hash), force=True)
    transaction_service.on_block_cleaned_up(message_hash)
    logger.statistics(
        {
            "type": "MemoryCleanup",
            "event": "CacheStateAfterBlockCleanup",
            "message_hash": repr(message_hash),
            "data": transaction_service.get_cache_state_json()
        }
    )



import datetime
from typing import NamedTuple, Optional, List

from bxcommon.utils.object_hash import Sha256Hash


class BlockInfo(NamedTuple):
    block_hash: Sha256Hash
    short_ids: List[int]
    start_datetime: datetime.datetime
    end_datetime: datetime.datetime
    duration_ms: float
    txn_count: Optional[int]
    compressed_block_hash: Optional[str]
    prev_block_hash: Optional[str]
    original_size: Optional[float]
    compressed_size: Optional[float]
    compression_rate: Optional[float]
    ignored_short_ids: List[Optional[int]]



from dataclasses import dataclass
from typing import Optional

from bxcommon.utils.blockchain_utils.eth import eth_common_constants


@dataclass
class BlockchainPeerInfo:
    ip: str
    port: int
    node_public_key: Optional[str] = None
    blockchain_protocol_version: int = eth_common_constants.ETH_PROTOCOL_VERSION
    connection_established: bool = False

    def __repr__(self):
        return f"BlockchainPeerInfo(ip address: {self.ip}, " \
               f"port: {self.port}, " \
               f"node public key: {self.node_public_key}, " \
               f"blockchain protocol version: {self.blockchain_protocol_version})"

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, BlockchainPeerInfo)
            and other.port == self.port
            and other.ip == self.ip
        )

    def __hash__(self):
        return hash(f"{self.ip}:{self.port}")



from bxcommon import constants
from bxcommon.messages.bloxroute.bloxroute_message_factory import bloxroute_message_factory
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.protocol_version import PROTOCOL_VERSION
from bxcommon.messages.bloxroute.v22 import message_converter_factory_v22, bloxroute_message_factory_v22
from bxcommon.messages.bloxroute.v6.bloxroute_message_factory_v6 import bloxroute_message_factory_v6
from bxcommon.messages.bloxroute.v6.message_converter_factory_v6 import message_converter_factory_v6
from bxcommon.messages.bloxroute.v7.bloxroute_message_factory_v7 import bloxroute_message_factory_v7
from bxcommon.messages.bloxroute.v7.message_converter_factory_v7 import message_converter_factory_v7
from bxcommon.messages.bloxroute.v8.bloxroute_message_factory_v8 import bloxroute_message_factory_v8
from bxcommon.messages.bloxroute.v8.message_converter_factory_v8 import message_converter_factory_v8
from bxcommon.messages.bloxroute.v9.bloxroute_message_factory_v9 import bloxroute_message_factory_v9
from bxcommon.messages.bloxroute.v9.message_converter_factory_v9 import message_converter_factory_v9
from bxcommon.messages.bloxroute.v10.bloxroute_message_factory_v10 import bloxroute_message_factory_v10
from bxcommon.messages.bloxroute.v10.message_converter_factory_v10 import message_converter_factory_v10
from bxcommon.messages.bloxroute.v11.bloxroute_message_factory_v11 import bloxroute_message_factory_v11
from bxcommon.messages.bloxroute.v11.message_converter_factory_v11 import message_converter_factory_v11
from bxcommon.messages.bloxroute.v12.bloxroute_message_factory_v12 import bloxroute_message_factory_v12
from bxcommon.messages.bloxroute.v12.message_converter_factory_v12 import message_converter_factory_v12
from bxcommon.messages.bloxroute.v13.bloxroute_message_factory_v13 import bloxroute_message_factory_v13
from bxcommon.messages.bloxroute.v13.message_converter_factory_v13 import message_converter_factory_v13
from bxcommon.messages.bloxroute.v14.bloxroute_message_factory_v14 import bloxroute_message_factory_v14
from bxcommon.messages.bloxroute.v14.message_converter_factory_v14 import message_converter_factory_v14
from bxcommon.messages.bloxroute.v15.bloxroute_message_factory_v15 import bloxroute_message_factory_v15
from bxcommon.messages.bloxroute.v15.message_converter_factory_v15 import message_converter_factory_v15
from bxcommon.messages.bloxroute.v16.bloxroute_message_factory_v16 import bloxroute_message_factory_v16
from bxcommon.messages.bloxroute.v16.message_converter_factory_v16 import message_converter_factory_v16
from bxcommon.messages.bloxroute.v17.bloxroute_message_factory_v17 import bloxroute_message_factory_v17
from bxcommon.messages.bloxroute.v17.message_converter_factory_v17 import message_converter_factory_v17
from bxcommon.messages.bloxroute.v18.message_converter_factory_v18 import message_converter_factory_v18
from bxcommon.messages.bloxroute.v18.bloxroute_message_factory_v18 import bloxroute_message_factory_v18
from bxcommon.messages.bloxroute.v19.bloxroute_message_factory_v19 import bloxroute_message_factory_v19
from bxcommon.messages.bloxroute.v19.message_converter_factory_v19 import message_converter_factory_v19
from bxcommon.messages.bloxroute.v20.message_converter_factory_v20 import message_converter_factory_v20
from bxcommon.messages.bloxroute.v20.bloxroute_message_factory_v20 import bloxroute_message_factory_v20
from bxcommon.messages.bloxroute.v21.message_converter_factory_v21 import message_converter_factory_v21
from bxcommon.messages.bloxroute.v21.bloxroute_message_factory_v21 import bloxroute_message_factory_v21
from bxcommon.messages.versioning.abstract_version_manager import AbstractVersionManager


class _BloxrouteVersionManager(AbstractVersionManager):
    CURRENT_PROTOCOL_VERSION = PROTOCOL_VERSION
    MIN_SUPPORTED_PROTOCOL_VERSION = 6
    VERSION_MESSAGE_MAIN_LENGTH = constants.VERSIONED_HELLO_MSG_MIN_PAYLOAD_LEN
    _PROTOCOL_TO_CONVERTER_FACTORY_MAPPING = {
        6: message_converter_factory_v6,
        7: message_converter_factory_v7,
        8: message_converter_factory_v8,
        9: message_converter_factory_v9,
        10: message_converter_factory_v10,
        11: message_converter_factory_v11,
        12: message_converter_factory_v12,
        13: message_converter_factory_v13,
        14: message_converter_factory_v14,
        15: message_converter_factory_v15,
        16: message_converter_factory_v16,
        17: message_converter_factory_v17,
        18: message_converter_factory_v18,
        19: message_converter_factory_v19,
        20: message_converter_factory_v20,
        21: message_converter_factory_v21,
        22: message_converter_factory_v22
    }
    _PROTOCOL_TO_FACTORY_MAPPING = {
        6: bloxroute_message_factory_v6,
        7: bloxroute_message_factory_v7,
        8: bloxroute_message_factory_v8,
        9: bloxroute_message_factory_v9,
        10: bloxroute_message_factory_v10,
        11: bloxroute_message_factory_v11,
        12: bloxroute_message_factory_v12,
        13: bloxroute_message_factory_v13,
        14: bloxroute_message_factory_v14,
        15: bloxroute_message_factory_v15,
        16: bloxroute_message_factory_v16,
        17: bloxroute_message_factory_v17,
        18: bloxroute_message_factory_v18,
        19: bloxroute_message_factory_v19,
        20: bloxroute_message_factory_v20,
        21: bloxroute_message_factory_v21,
        22: bloxroute_message_factory_v22,
        23: bloxroute_message_factory
    }

    def __init__(self) -> None:
        super(_BloxrouteVersionManager, self).__init__()
        self.protocol_to_factory_mapping = self._PROTOCOL_TO_FACTORY_MAPPING
        self.protocol_to_converter_factory_mapping = self._PROTOCOL_TO_CONVERTER_FACTORY_MAPPING
        self.version_message_command = BloxrouteMessageType.HELLO


bloxroute_version_manager = _BloxrouteVersionManager()



from typing import Generic, TypeVar, Callable, Any, Dict, Iterator

T = TypeVar("T")


class ObjectPriority(Generic[T]):
    """
    an object wrapper that provide a priority value to it
    the priority value must be a sortable type.
    I.E: implement both __eq__ and __lt__
    """

    def __init__(self, get_priority: Callable[[T], Any], obj: T) -> None:
        self._get_priority = get_priority
        self._obj = obj

    def __repr__(self):
        return f"{self.__class__.__name__}<{self._obj}>"

    def get_priority(self) -> Any:
        return self._get_priority(self._obj)

    def get_object(self) -> T:
        return self._obj


class PriorityQueue(Generic[T]):
    """
     a collection class that sort its items according to their priority.
     items needs to be wrapped with a priority object (ObjectPriority) first, using the add method.
     items needs to be removed from the queue regardless weather they where popped from the queue or not
     in order to remove the priority wrapper they where added with (aka "object tracker").
     sorting by priority should be performed by explicitly call the update_priority method.
     """

    def __init__(self, is_reversed: bool = True) -> None:
        """
        initialize the priority queue collection object.
        :param is_reversed: indicate if the sorting should be performed in revered order.
        """
        self._items_dict: Dict[T, ObjectPriority] = {}
        self._items_tracker: Dict[T, ObjectPriority] = {}
        self._is_reversed = is_reversed

    def __repr__(self):
        return f"{self.__class__.__name__}<{list(self._items_dict.values())}>"

    def __len__(self) -> int:
        return len(self._items_dict)

    def __bool__(self) -> bool:
        return len(self) > 0

    def __iter__(self) -> Iterator[T]:
        for obj in self._items_dict:
            yield obj

    def add(self, item: ObjectPriority[T]) -> None:
        """
        adds a priority item to the queue.
        will track the underlying object until remove is called.
        :param item: see [ObjectPriority] docs
        :raise ValueError: if the item was already added (should call push instead)
        """
        if item.get_object() not in self._items_tracker:
            self._items_tracker[item.get_object()] = item
            self._items_dict[item.get_object()] = item
        else:
            raise ValueError(f"priority item {item} was already added")

    def push(self, obj: T) -> None:
        """
        push an object back to the priority queue
        :param obj: the object to be pushed
        :raise ValueError: if the object was never added to the queue
        """
        if obj in self._items_tracker:
            priority_item = self._items_tracker[obj]
            if obj not in self._items_dict:
                self._items_dict[obj] = priority_item
        else:
            raise self._missing_object_error(obj)

    def remove(self, obj: T) -> None:
        """
        remove an object from the queue
        :param obj: the object to be removed
        """
        try:
            del self._items_tracker[obj]
            del self._items_dict[obj]
        except KeyError:
            pass

    def pop(self) -> T:
        """
        pop an item from the queue, but keep track of its priority wrapper.
        the object will not be included in the priority queue but still needs to be removed from the tracker
        :return: the first item in the priority queue
        """
        obj, _ = self._items_dict.popitem()
        return obj

    def try_remove_from_queue(self, obj: T) -> None:
        """
        attempts to remove an item from the priority queue (not from the tracker).
        :param obj: the object to be removed.
        :raise ValueError: if the object was never added to the queue
        """
        if obj in self._items_tracker:
            try:
                del self._items_dict[obj]
            except KeyError:
                pass
        else:
            raise self._missing_object_error(obj)

    def update_priority(self) -> None:
        """
        updating the priority queue by fetching each object's priority and sort by it.
        sorting is being done according to the is_reversed flag passed in the constructor.
        In case of True, for example: [5, 33, 1, 0, 2, 4] -> [33, 5, 4, 2, 1, 0]
        this method has at least O(n log n) complexity and should be called as less frequent as possible.
        """
        sorted_items = sorted(self._items_dict.values(), key=lambda p: p.get_priority(), reverse=self._is_reversed)
        new_items_dict = {}
        for item in sorted_items:
            new_items_dict[item.get_object()] = item

        self._items_dict = new_items_dict

    def _missing_object_error(self, obj) -> ValueError:
        return ValueError(f"object {obj} is missing in items tracker {self._items_tracker}")



from bxgateway.messages.ont.ont_normal_consensus_message_converter import OntNormalConsensusMessageConverter


def create_ont_consensus_message_converter(magic, opts):
    if opts.use_extensions or opts.import_extensions:
        from bxgateway.messages.ont.ont_extension_consensus_message_converter import OntExtensionConsensusMessageConverter

    if opts.use_extensions:
        return OntExtensionConsensusMessageConverter(magic)
    else:
        return OntNormalConsensusMessageConverter(magic)



import typing
from datetime import datetime

from bxcommon.connections.connection_type import ConnectionType
from bxutils import logging
from bxgateway import log_messages

from bxcommon.utils import convert
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.stats import stats_format
from bxcommon.utils.stats.block_stat_event_type import BlockStatEventType
from bxcommon.utils.stats.block_statistics_service import block_stats

from bxgateway.connections.btc.btc_node_connection import BtcNodeConnection
from bxgateway.messages.btc.abstract_btc_message_converter import CompactBlockCompressionResult
from bxgateway.messages.btc.block_btc_message import BlockBtcMessage
from bxgateway.messages.btc.block_transactions_btc_message import BlockTransactionsBtcMessage
from bxgateway.messages.btc.compact_block_btc_message import CompactBlockBtcMessage
from bxgateway.messages.btc.inventory_btc_message import GetDataBtcMessage, InventoryType
from bxgateway.services.block_processing_service import BlockProcessingService
from bxgateway.utils.errors.message_conversion_error import MessageConversionError

logger = logging.get_logger(__name__)


class BtcBlockProcessingService(BlockProcessingService):

    def process_compact_block(
            self, block_message: CompactBlockBtcMessage, connection: BtcNodeConnection
    ) -> CompactBlockCompressionResult:
        """
        Process compact block for processing on timeout if hold message received.
        If no hold exists, compress and broadcast block immediately.
        :param block_message: compact block message to process
        :param connection: receiving connection (AbstractBlockchainConnection)
        """
        block_hash = block_message.block_hash()
        parse_result = self._node.message_converter.compact_block_to_bx_block(  # pyre-ignore
            block_message,
            self._node.get_tx_service()
        )
        block_info = parse_result.block_info
        if parse_result.success:
            block_stats.add_block_event_by_block_hash(
                block_hash,
                BlockStatEventType.COMPACT_BLOCK_COMPRESSED_SUCCESS,
                network_num=connection.network_num,
                start_date_time=block_info.start_datetime,
                end_date_time=block_info.end_datetime,
                duration=block_info.duration_ms / 1000,
                success=parse_result.success,
                txs_count=parse_result.block_info.txn_count,
                prev_block_hash=parse_result.block_info.prev_block_hash,
                original_size=block_info.original_size,
                more_info="Compression: {}->{} bytes, {}, {}; Tx count: {}".format(
                    block_info.original_size,
                    block_info.compressed_size,
                    stats_format.percentage(block_info.compression_rate),
                    stats_format.duration(block_info.duration_ms),
                    block_info.txn_count)
            )
            self._node.block_cleanup_service.on_new_block_received(block_hash, block_message.prev_block_hash())
            self._process_and_broadcast_compressed_block(
                parse_result.bx_block,
                connection,
                parse_result.block_info,
                block_hash
            )
        else:
            missing_indices = parse_result.missing_indices
            missing_indices_count = 0 if missing_indices is None else len(missing_indices)
            start_datetime = block_info.start_datetime
            end_datetime = datetime.utcnow()
            duration = (end_datetime - start_datetime).total_seconds()
            block_stats.add_block_event_by_block_hash(
                block_hash,
                BlockStatEventType.COMPACT_BLOCK_COMPRESSED_FAILED,
                network_num=connection.network_num,
                start_date_time=start_datetime,
                end_date_time=end_datetime,
                duration=duration,
                success=parse_result.success,
                missing_short_id_count=missing_indices_count,
                more_info="{:.2f}ms".format(
                    duration * 1000
                )
            )
            logger.warning(log_messages.UNKNOWN_SHORT_IDS,
                           missing_indices_count)
        return parse_result

    def process_compact_block_recovery(
            self,
            msg: BlockTransactionsBtcMessage,
            failure_result: CompactBlockCompressionResult,
            connection: BtcNodeConnection
    ) -> None:
        """
        Process compact block recovery .
        If no hold exists, compress and broadcast block immediately.
        """
        block_hash = msg.block_hash()

        for txn in msg.transactions():
            failure_result.recovered_transactions.append(txn)

        try:
            recovery_result = self._node.message_converter.recovered_compact_block_to_bx_block(  # pyre-ignore
                failure_result
            )
        except MessageConversionError as e:
            block_stats.add_block_event_by_block_hash(
                e.msg_hash,
                BlockStatEventType.BLOCK_CONVERSION_FAILED,
                network_num=connection.network_num,
                conversion_type=e.conversion_type.value
            )
            logger.warning(log_messages.COMPACT_BLOCK_PROCESSING_FAIL,
                           e.msg_hash, e)
            get_data_msg = GetDataBtcMessage(
                magic=msg.magic(),
                inv_vects=[(InventoryType.MSG_BLOCK, msg.block_hash())]
            )
            connection.enqueue_msg(get_data_msg)
            return
        block_info = recovery_result.block_info

        if recovery_result.success:
            block_stats.add_block_event_by_block_hash(
                block_hash,
                BlockStatEventType.COMPACT_BLOCK_RECOVERY_SUCCESS,
                network_num=connection.network_num,
                start_date_time=block_info.start_datetime,
                end_date_time=block_info.end_datetime,
                duration=block_info.duration_ms / 1000,
                success=recovery_result.success,
                recoverd_txs_count=len(msg.transactions()),
                txs_count=recovery_result.block_info.txn_count,
                prev_block_hash=recovery_result.block_info.prev_block_hash,
                more_info="{:.2f}ms, {:f}".format(
                    block_info.duration_ms,
                    len(msg.transactions())
                )
            )
            prev_block = Sha256Hash(convert.hex_to_bytes(block_info.prev_block_hash))
            self._node.block_cleanup_service.on_new_block_received(block_hash, prev_block)
            self._process_and_broadcast_compressed_block(
                recovery_result.bx_block,
                connection,
                recovery_result.block_info,
                msg.block_hash()
            )
        else:
            start_datetime = block_info.start_datetime
            end_datetime = datetime.utcnow()
            duration = (end_datetime - start_datetime).total_seconds()
            block_stats.add_block_event_by_block_hash(
                block_hash,
                BlockStatEventType.COMPACT_BLOCK_RECOVERY_FAILED,
                network_num=connection.network_num,
                start_date_time=start_datetime,
                end_date_time=end_datetime,
                duration=duration,
                success=recovery_result.success,
                recoverd_txs_count=len(msg.transactions()),
                more_info="{:.2f}ms, {:f}".format(
                    duration * 1000,
                    len(msg.transactions())
                )
            )
            logger.warning(log_messages.COMPACT_BLOCK_RECOVERY_FAIL, msg.block_hash())
            get_data_msg = GetDataBtcMessage(
                magic=msg.magic(),
                inv_vects=[(InventoryType.MSG_BLOCK, msg.block_hash())]
            )
            connection.enqueue_msg(get_data_msg)

    def _on_block_decompressed(self, block_msg):
        msg = typing.cast(BlockBtcMessage, block_msg)
        self._node.block_cleanup_service.on_new_block_received(msg.block_hash(), msg.prev_block_hash())



import bxgateway.messages.btc.btc_message_converter_factory as converter_factory
from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxgateway.connections.abstract_gateway_blockchain_connection import AbstractGatewayBlockchainConnection
from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode
from bxgateway.connections.abstract_relay_connection import AbstractRelayConnection
from bxgateway.connections.btc.btc_node_connection import BtcNodeConnection
from bxgateway.connections.btc.btc_relay_connection import BtcRelayConnection
from bxgateway.connections.btc.btc_remote_connection import BtcRemoteConnection
from bxgateway.services.abstract_block_cleanup_service import AbstractBlockCleanupService
from bxgateway.services.btc.btc_block_processing_service import BtcBlockProcessingService
from bxgateway.services.btc.btc_block_queuing_service import BtcBlockQueuingService
from bxgateway.services.btc.btc_normal_block_cleanup_service import BtcNormalBlockCleanupService
from bxgateway.services.push_block_queuing_service import PushBlockQueuingService
from bxgateway.testing.btc_lossy_relay_connection import BtcLossyRelayConnection
from bxgateway.testing.test_modes import TestModes
from bxutils.services.node_ssl_service import NodeSSLService


class BtcGatewayNode(AbstractGatewayNode):
    def __init__(self, opts, node_ssl_service: NodeSSLService):
        super(BtcGatewayNode, self).__init__(opts, node_ssl_service)

        self.block_processing_service = BtcBlockProcessingService(self)

        self.message_converter = converter_factory.create_btc_message_converter(
            self.opts.blockchain_net_magic,
            self.opts
        )

    def build_blockchain_connection(
        self, socket_connection: AbstractSocketConnectionProtocol
    ) -> AbstractGatewayBlockchainConnection:
        return BtcNodeConnection(socket_connection, self)

    def build_relay_connection(self, socket_connection: AbstractSocketConnectionProtocol) -> AbstractRelayConnection:
        if TestModes.DROPPING_TXS in self.opts.test_mode:
            cls = BtcLossyRelayConnection
        else:
            cls = BtcRelayConnection

        relay_connection = cls(socket_connection, self)
        return relay_connection

    def build_remote_blockchain_connection(
        self, socket_connection: AbstractSocketConnectionProtocol
    ) -> AbstractGatewayBlockchainConnection:
        return BtcRemoteConnection(socket_connection, self)

    def build_block_queuing_service(
        self,
        connection: AbstractGatewayBlockchainConnection
    ) -> PushBlockQueuingService:
        return BtcBlockQueuingService(self, connection)

    def build_block_cleanup_service(self) -> AbstractBlockCleanupService:
        if self.opts.use_extensions:
            from bxgateway.services.btc.btc_extension_block_cleanup_service import BtcExtensionBlockCleanupService
            block_cleanup_service = BtcExtensionBlockCleanupService(self, self.network_num)
        else:
            block_cleanup_service = BtcNormalBlockCleanupService(self, self.network_num)
        return block_cleanup_service



from bxcommon.utils.stats.stat_event_type_settings import StatEventTypeSettings


class ConnectionStateEventType:

    REGISTER_PENDING = StatEventTypeSettings("ConnectionRegisterPending")
    BUCKET_DEPLETED = StatEventTypeSettings("ConnectionBucketDepleted")
    TIMED_OUT = StatEventTypeSettings("ConnectionTimedOut")
    RE_REGISTERED = StatEventTypeSettings("ConnectionReRegistered")
    GROUP_CHANGED = StatEventTypeSettings("ConnectionDDoSGroupChanged")
    CONNECTION_DISCONNECTED = StatEventTypeSettings("ConnectionDisconnected")



from typing import Optional

from bxcommon import constants
from bxcommon.messages.bloxroute.abstract_broadcast_message import AbstractBroadcastMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.utils import crypto
from bxcommon.utils.object_hash import Sha256Hash
from bxutils.logging.log_level import LogLevel


class KeyMessage(AbstractBroadcastMessage):
    MESSAGE_TYPE = BloxrouteMessageType.KEY
    PAYLOAD_LENGTH = AbstractBroadcastMessage.PAYLOAD_LENGTH + crypto.KEY_SIZE

    def __init__(self, message_hash: Optional[Sha256Hash] = None, network_num: Optional[int] = None,
                 source_id: str = "", key: Optional[bytearray] = None, buf: Optional[bytearray] = None):
        self._key = None
        self._block_id = None

        super().__init__(message_hash, network_num, source_id, buf)

        if buf is None:
            # pyre-fixme[6]: Expected `Sized` for 1st param but got
            #  `Optional[bytearray]`.
            if len(key) != crypto.KEY_SIZE:
                raise ValueError(f"Key must be of size {crypto.KEY_SIZE}")

            off = self.HEADER_LENGTH + AbstractBroadcastMessage.PAYLOAD_LENGTH - constants.CONTROL_FLAGS_LEN
            self.buf[off:off + crypto.KEY_SIZE] = key

    def log_level(self):
        return LogLevel.DEBUG

    def block_hash(self) -> Sha256Hash:
        return self.message_hash()

    def key(self) -> memoryview:
        if self._key is None:
            off = self.HEADER_LENGTH + AbstractBroadcastMessage.PAYLOAD_LENGTH - constants.CONTROL_FLAGS_LEN
            self._key = self._memoryview[off:off + crypto.KEY_SIZE]

        # pyre-fixme[7]: Expected `memoryview` but got `None`.
        return self._key

    def __repr__(self):
        return f"KeyMessage<network_num: {self.network_num()}, block_id: {self.message_id()}"



import blxr_rlp as rlp
from bxcommon.utils.blockchain_utils.eth import eth_common_constants


class Eip8Auth(rlp.Serializable):
    fields = [
        ("signature", rlp.sedes.Binary.fixed_length(eth_common_constants.SIGNATURE_LEN)),
        ("public_key", rlp.sedes.Binary.fixed_length(eth_common_constants.PUBLIC_KEY_LEN)),
        ("nonce", rlp.sedes.Binary.fixed_length(eth_common_constants.AUTH_NONCE_LEN)),
        ("version", rlp.sedes.big_endian_int)
    ]


class Eip8AuthAck(rlp.Serializable):
    fields = [
        ("public_key", rlp.sedes.Binary.fixed_length(eth_common_constants.PUBLIC_KEY_LEN)),
        ("nonce", rlp.sedes.Binary.fixed_length(eth_common_constants.AUTH_NONCE_LEN)),
        ("version", rlp.sedes.big_endian_int)
    ]



import struct
from typing import Optional

from bxgateway import ont_constants
from bxgateway.messages.ont.ont_message import OntMessage
from bxgateway.messages.ont.ont_message_type import OntMessageType


class VerAckOntMessage(OntMessage):
    MESSAGE_TYPE = OntMessageType.VERACK

    def __init__(self, magic: Optional[int] = None, is_consensus: Optional[bool] = None,
                 buf: Optional[bytearray] = None):
        if buf is None:
            buf = bytearray(ont_constants.ONT_HDR_COMMON_OFF + ont_constants.ONT_CHAR_LEN)
            self.buf = buf

            off = ont_constants.ONT_HDR_COMMON_OFF
            struct.pack_into("<?", buf, off, is_consensus)
            off += ont_constants.ONT_CHAR_LEN

            super().__init__(magic, self.MESSAGE_TYPE, off - ont_constants.ONT_HDR_COMMON_OFF, buf)
        else:
            self.buf = buf
            self._memoryview = memoryview(buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None
            self._is_consensus = None

    def is_consensus(self) -> bool:
        if self._is_consensus is None:
            off = ont_constants.ONT_HDR_COMMON_OFF
            self._is_consensus, = struct.unpack_from("<?", self.buf, off)

        is_consensus = self._is_consensus
        assert isinstance(is_consensus, bool)
        return is_consensus



import functools
import os
import ssl
import urllib.request
import urllib.response
from asyncio import Transport
from datetime import datetime, time, date
from ssl import SSLSocket
from typing import Callable, Iterable, Union, IO

from cryptography import x509
from cryptography.hazmat import backends
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKeyWithSerialization, \
    EllipticCurvePrivateKey
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.x509 import Certificate, CertificateSigningRequest, \
    CertificateSigningRequestBuilder, SubjectAlternativeName, ExtensionNotFound, \
    CertificateBuilder, AuthorityKeyIdentifier
from cryptography.x509.extensions import KeyUsage, UnrecognizedExtension, BasicConstraints

# pylint: disable=no-name-in-module
# pyre-fixme[21]: Could not find `loop`.
from uvloop.loop import TCPTransport

from bxutils import constants
from bxutils.ssl import ssl_serializer

KEY_FILE_FORMAT = constants.SSL_KEY_FILE_FORMAT
CERT_FILE_FORMAT = constants.SSL_CERT_FILE_FORMAT


def create_csr(key: EllipticCurvePrivateKey, template_cert: Certificate) -> CertificateSigningRequest:
    """
    Create a Certificate Signing Request (CSR)
    matching to a corresponding template_cert and sign it with a private key.
    :param key: the private key to sign the CSR
    :param template_cert: a template for retrieving the subject information.
    :return: CSR (https://en.wikipedia.org/wiki/Certificate_signing_request)
    """
    csr_builder = CertificateSigningRequestBuilder().subject_name(
        template_cert.subject
    )
    try:
        subject_alt_name = template_cert.extensions.get_extension_for_class(SubjectAlternativeName)
        csr_builder = csr_builder.add_extension(
            subject_alt_name.value,
            critical=False
        )
    except ExtensionNotFound:
        pass

    return csr_builder.sign(key, SHA256(), backends.default_backend())


def generate_key() -> EllipticCurvePrivateKeyWithSerialization:
    """
    Generate an EC (https://en.wikipedia.org/wiki/Elliptic-curve_cryptography) private key
    to be used for signing an SSL certificate or CSR.
    :return: EC private key
    """
    return ec.generate_private_key(ec.SECP384R1(), backends.default_backend())


def store(
        path: str,
        data: bytes,
        stream_factory: Callable[[str], IO] = functools.partial(open, mode="wb")
) -> None:
    """
    Stores raw certificate data (in PEM format)
    :param path: the file path in which the data will be written to.
    :param data: the PEM encoded certificate data
    :param stream_factory: a handler callback to open a file stream (to make this method more testable)
    """
    with stream_factory(path) as stream:
        stream.write(data)


def store_key(
        key: EllipticCurvePrivateKeyWithSerialization,
        folder_path: str,
        name: str,
        stream_factory: Callable[[str], IO] = functools.partial(open, mode="wb")
) -> str:
    """
    Stores an EC private key.
    :param key: the private key to store
    :param folder_path: the ssl folder path
    :param name: the name prefix for the key file (see: KEY_FILE_FORMAT)
    :param stream_factory: a handler callback to open a file stream (to make this method more testable)
    :return: the full path of the key file
    """
    file_path = os.path.join(folder_path, KEY_FILE_FORMAT.format(name))
    store(
        file_path,
        ssl_serializer.serialize_key(key).encode("utf-8"),
        stream_factory
    )
    return file_path


def store_certificate(
        cert: Certificate,
        folder_path: str,
        name: str,
        stream_factory: Callable[[str], IO] = functools.partial(open, mode="wb")
) -> str:
    """
    Stores an SSL certificate.
    :param cert: the certificate to store
    :param folder_path: the ssl folder path
    :param name: the name prefix for the certificate file (see: CERT_FILE_FORMAT)
    :param stream_factory: a handler callback to open a file stream (to make this method more testable)
    :return: the full path of the certificate file
    """
    file_path = os.path.join(folder_path, CERT_FILE_FORMAT.format(name))
    store(
        file_path,
        ssl_serializer.serialize_cert(cert).encode("utf-8"),
        stream_factory
    )
    return file_path


def fetch_file(url: str) -> IO:
    """
    Fetch a PEM encoded file from a URL.
    :param url: the file URL
    :return: stream like object
    """
    return urllib.request.urlopen(url)


def fetch_cert(url: str) -> Certificate:
    """
    Fetch a certificate from a URL.
    :param url: the URL to the certificate file
    :return: a certificate object
    """
    with fetch_file(url) as cert_file:
        return ssl_serializer.deserialize_cert(cert_file.read())


def fetch_key(url: str) -> EllipticCurvePrivateKeyWithSerialization:
    """
    Fetch a private key from a URL.
    :param url: the URL to the private key file
    :return: a private key object
    """
    with fetch_file(url) as key_file:
        return ssl_serializer.deserialize_key(key_file.read())


def get_socket_cert(ssl_socket: Union[SSLSocket, ssl.SSLObject]) -> Certificate:
    """
    Obtain a peer certificate from an SSL socket.
    :param ssl_socket: the SSL socket object
    :return: a certificate object
    """
    der_cert = ssl_socket.getpeercert(True)
    pem_cert = ssl.DER_cert_to_PEM_cert(der_cert)  # pyre-ignore
    return ssl_serializer.deserialize_cert(pem_cert)


# pyre-fixme[11]: Annotation `TCPTransport` is not defined as a type.
def get_transport_cert(transport: Union[Transport, TCPTransport]) -> Certificate:
    """
    Obtain a peer certificate from a Transport socket wrapper
    :param transport: the SSL socket transport wrapper
    :return: a certificate object
    :raise: ValueError if the transport doesn't wrap an SSL socket
    """
    ssl_socket = transport.get_extra_info("ssl_object")
    if isinstance(ssl_socket, (SSLSocket, ssl.SSLObject)):
        return get_socket_cert(ssl_socket)
    else:
        raise ValueError("transport does not wrap an ssl socket")


def sign_csr(
    csr: CertificateSigningRequest,
    ca_cert: Certificate,
    key: EllipticCurvePrivateKey,
    expiration_date: date,
    custom_extensions: Iterable[Union[KeyUsage, UnrecognizedExtension, BasicConstraints]]
) -> Certificate:
    """
    Sign a CSR with CA credentials.
    :param csr: the CSR
    :param ca_cert: the CA certificate
    :param key: the CA private key
    :param expiration_date: expiration date
    :param custom_extensions: custom extensions to be added to the certificate
    :return: a certificate object
    """
    issuer = ca_cert.subject
    now = datetime.utcnow()
    cert_builder = CertificateBuilder().issuer_name(issuer).subject_name(csr.subject).public_key(
        csr.public_key()
    ).serial_number(
        x509.random_serial_number()
    ).not_valid_before(now).not_valid_after(
        datetime.combine(expiration_date, time(), None)
    ).add_extension(
        extension=AuthorityKeyIdentifier.from_issuer_public_key(ca_cert.public_key()),
        critical=False
    )
    try:
        cert_builder = cert_builder.add_extension(
            csr.extensions.get_extension_for_class(SubjectAlternativeName).value, critical=False
        )
    except ExtensionNotFound:
        pass
    for extension in custom_extensions:
        if isinstance(extension, UnrecognizedExtension):
            critical = False
        else:
            critical = True
        # pyre-fixme[6]: Expected `ExtensionType` for 1st param but got
        #  `Union[BasicConstraints, KeyUsage, UnrecognizedExtension]`.
        cert_builder = cert_builder.add_extension(extension, critical=critical)
    return cert_builder.sign(key, SHA256(), backends.default_backend())



import struct
from typing import Optional

from bxcommon import constants
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.abstract_broadcast_message import AbstractBroadcastMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.utils import crypto
from bxcommon.utils.buffers.input_buffer import InputBuffer
from bxcommon.utils.object_hash import Sha256Hash, AbstractObjectHash
from bxutils.logging.log_level import LogLevel


class BroadcastMessageV8(AbstractBroadcastMessage):
    MESSAGE_TYPE = BloxrouteMessageType.BROADCAST

    def __init__(self, message_hash: Optional[Sha256Hash] = None, network_num: Optional[int] = None,
                 source_id: str = "", is_encrypted: bool = False, blob: Optional[bytearray] = None,
                 buf: Optional[bytearray] = None):
        self._block_id = None
        self._is_encrypted = None
        self._blob = None

        # override payload length for variable length message
        if blob:
            # pylint: disable=invalid-name
            self.PAYLOAD_LENGTH = (
                AbstractBroadcastMessage.PAYLOAD_LENGTH
                + constants.BLOCK_ENCRYPTED_FLAG_LEN
                + len(blob)
            )
        super().__init__(message_hash, network_num, source_id, buf)

        if buf is None:
            # minus control flag
            off = self.HEADER_LENGTH + AbstractBroadcastMessage.PAYLOAD_LENGTH - constants.CONTROL_FLAGS_LEN

            struct.pack_into("?", self.buf, off, is_encrypted)
            off += constants.BLOCK_ENCRYPTED_FLAG_LEN

            # pyre-fixme[6]: Expected `Sized` for 1st param but got
            #  `Optional[bytearray]`.
            self.buf[off:off + len(blob)] = blob

    def log_level(self) -> LogLevel:
        return LogLevel.DEBUG

    def is_encrypted(self) -> bool:
        if self._is_encrypted is None:
            off = self.HEADER_LENGTH + AbstractBroadcastMessage.PAYLOAD_LENGTH - constants.CONTROL_FLAGS_LEN
            self._is_encrypted, = struct.unpack_from("?", self.buf, off)
        # pyre-fixme[7]: Expected `bool` but got `None`.
        return self._is_encrypted

    def block_hash(self) -> AbstractObjectHash:
        return self.message_hash()

    def blob(self) -> memoryview:
        if self._blob is None:
            off = self.HEADER_LENGTH + AbstractBroadcastMessage.PAYLOAD_LENGTH + constants.BLOCK_ENCRYPTED_FLAG_LEN - \
                  constants.CONTROL_FLAGS_LEN
            self._blob = self._memoryview[off:self.HEADER_LENGTH + self.payload_len() - constants.CONTROL_FLAGS_LEN]

        # pyre-fixme[7]: Expected `memoryview` but got `None`.
        return self._blob

    @classmethod
    def peek_network_num(cls, input_buffer: InputBuffer) -> int:
        off = AbstractBloxrouteMessage.HEADER_LENGTH + crypto.SHA256_HASH_LEN

        if input_buffer.length < off + constants.NETWORK_NUM_LEN:
            raise ValueError("Not enough bytes to peek network number.")

        network_num_bytes = input_buffer.peek_message(off + constants.NETWORK_NUM_LEN)

        network_num, = struct.unpack_from("<L", network_num_bytes, off)

        return network_num

    def __repr__(self) -> str:
        return "BroadcastMessage<network_num: {}, block_id: {}, blob_length: {}, is_encrypted: {}, source_id: {}>" \
            .format(self.network_num(), self.message_id(), len(self.blob()), self.is_encrypted(),
                    self.source_id_as_str())



from typing import Dict, Any

import blxr_rlp as rlp

from bxcommon.utils import convert
from bxcommon.utils.blockchain_utils.eth import eth_common_utils, eth_common_constants
from bxcommon.utils.object_hash import Sha256Hash


class BlockHeader(rlp.Serializable):
    FIXED_LENGTH_FIELD_OFFSET = 2 * eth_common_constants.BLOCK_HASH_LEN + eth_common_constants.ADDRESS_LEN + \
                                3 * eth_common_constants.MERKLE_ROOT_LEN + eth_common_constants.BLOOM_LEN + 9

    fields = [
        ("prev_hash", rlp.sedes.Binary.fixed_length(eth_common_constants.BLOCK_HASH_LEN)),
        ("uncles_hash", rlp.sedes.Binary.fixed_length(eth_common_constants.BLOCK_HASH_LEN)),
        ("coinbase", rlp.sedes.Binary.fixed_length(eth_common_constants.ADDRESS_LEN, allow_empty=True)),
        ("state_root", rlp.sedes.Binary.fixed_length(eth_common_constants.MERKLE_ROOT_LEN, allow_empty=True)),
        ("tx_list_root", rlp.sedes.Binary.fixed_length(eth_common_constants.MERKLE_ROOT_LEN, allow_empty=True)),
        ("receipts_root", rlp.sedes.Binary.fixed_length(eth_common_constants.MERKLE_ROOT_LEN, allow_empty=True)),
        ("bloom", rlp.sedes.BigEndianInt(eth_common_constants.BLOOM_LEN)),
        ("difficulty", rlp.sedes.big_endian_int),
        ("number", rlp.sedes.big_endian_int),
        ("gas_limit", rlp.sedes.big_endian_int),
        ("gas_used", rlp.sedes.big_endian_int),
        ("timestamp", rlp.sedes.big_endian_int),
        ("extra_data", rlp.sedes.binary),
        ("mix_hash", rlp.sedes.Binary.fixed_length(eth_common_constants.BLOCK_HASH_LEN)),
        ("nonce", rlp.sedes.Binary.fixed_length(eth_common_constants.BLOCK_NONCE_LEN)),
    ]

    prev_hash: bytearray
    uncles_hash: bytearray
    coinbase: bytearray
    state_root: bytearray
    tx_list_root: bytearray
    receipts_root: bytearray
    bloom: int = 0
    difficulty: int = 0
    number: int = 0
    gas_limit: int = 0
    gas_used: int = 0
    timestamp: int = 0
    extra_data: bytearray
    mix_hash: bytearray
    nonce: bytearray

    def __init__(self, *args, **kwargs):
        self.prev_hash = bytearray()
        self.uncles_hash = bytearray()
        self.coinbase = bytearray()
        self.state_root = bytearray()
        self.tx_list_root = bytearray()
        self.receipts_root = bytearray()
        self.extra_data = bytearray()
        self.mix_hash = bytearray()
        self.nonce = bytearray()

        super().__init__(*args, **kwargs)

    def __repr__(self) -> str:
        return f"EthBlockHeader<{self.hash_object()}>"

    # pylint: disable=arguments-differ
    @classmethod
    def deserialize(cls, serial, type_parsed: bool = False, **extra_kwargs):
        if type_parsed:
            return super().deserialize(serial, **extra_kwargs)

        if len(serial) == len(BlockHeader.fields):
            return super().deserialize(serial, **extra_kwargs)
        else:
            return LondonBlockHeader.deserialize(serial, type_parsed=True, **extra_kwargs)

    def get_field_value(self, field_name) -> Any:
        return getattr(self, field_name, None)

    def hash(self) -> bytes:
        """The binary block hash"""
        return eth_common_utils.keccak_hash(rlp.encode(self))

    def hash_object(self) -> Sha256Hash:
        return Sha256Hash(eth_common_utils.keccak_hash(rlp.encode(self)))

    def to_json(self) -> Dict[str, Any]:
        """
        Serializes data for publishing to the block feed.
        """
        return {
            "parent_hash": convert.bytes_to_hex_string_format(self.get_field_value('prev_hash')),
            "sha3_uncles": convert.bytes_to_hex_string_format(self.get_field_value('uncles_hash')),
            "miner": convert.bytes_to_hex_string_format(self.get_field_value("coinbase")),
            "state_root": convert.bytes_to_hex_string_format(self.get_field_value("state_root")),
            "transactions_root": convert.bytes_to_hex_string_format(self.get_field_value("tx_list_root")),
            "receipts_root": convert.bytes_to_hex_string_format(self.get_field_value("receipts_root")),
            "logs_bloom": hex(self.get_field_value("bloom")),
            "difficulty": hex(self.get_field_value("difficulty")),
            "number": hex(self.get_field_value("number")),
            "gas_limit": hex(self.get_field_value("gas_limit")),
            "gas_used": hex(self.get_field_value("gas_used")),
            "timestamp": hex(self.get_field_value("timestamp")),
            "extra_data": convert.bytes_to_hex_string_format(self.get_field_value("extra_data")),
            "mix_hash": convert.bytes_to_hex_string_format(self.get_field_value("mix_hash")),
            "nonce": convert.bytes_to_hex_string_format(self.get_field_value("nonce"))
        }


class LondonBlockHeader(BlockHeader):
    fields = [
        *BlockHeader.fields,
        ("base_fee_per_gas", rlp.sedes.big_endian_int),
    ]

    base_fee_per_gas: int = 0

    def __repr__(self) -> str:
        return f"EthLondonBlockHeader<{self.hash_object()}>"

    def to_json(self) -> Dict[str, Any]:
        block_header_json = super().to_json()
        block_header_json.update(
            {
                "base_fee_per_gas": self.get_field_value("base_fee_per_gas")
            }
        )
        return block_header_json


# pylint: disable=protected-access
# RLP hack to allow proper inheritance in sequential calls
LondonBlockHeader._sedes = rlp.sedes.List(sedes for _, sedes in LondonBlockHeader.fields)



from typing import Dict, Any

from bxcommon import constants
from bxcommon.rpc.rpc_errors import RpcInvalidParams
from bxcommon.rpc import rpc_constants
from bxcommon.feed.feed import Feed
from bxcommon.feed.subscriber import Subscriber
from bxcommon.utils.alarm_queue import AlarmQueue
from bxcommon.utils.expiring_set import ExpiringSet
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.feed.eth.eth_transaction_feed_entry import EthTransactionFeedEntry
from bxcommon.feed.eth.eth_raw_transaction import EthRawTransaction
from bxcommon.feed.eth import eth_filter_handlers
from bxutils import logging
from bxutils.logging.log_record_type import LogRecordType

logger = logging.get_logger()
logger_filters = logging.get_logger(LogRecordType.TransactionFiltering, __name__)

EXPIRATION_TIME_S = 5 * 60


class EthPendingTransactionFeed(Feed[EthTransactionFeedEntry, EthRawTransaction]):
    NAME = rpc_constants.ETH_PENDING_TRANSACTION_FEED_NAME
    FIELDS = [
        "tx_hash",
        "tx_contents",
        "tx_contents.nonce",
        "tx_contents.gas_price",
        "tx_contents.gas",
        "tx_contents.to",
        "tx_contents.max_priority_fee_per_gas",
        "tx_contents.max_fee_per_gas",
        "tx_contents.value",
        "tx_contents.input",
        "tx_contents.v",
        "tx_contents.r",
        "tx_contents.s",
        "tx_contents.from",
        "tx_contents.type",
        "local_region",
    ]
    FILTERS = {"value", "from", "to", "gas_price", "method_id", "max_priority_fee_per_gas", "max_fee_per_gas"}
    ALL_FIELDS = ["tx_hash", "tx_contents", "local_region"]

    published_transactions: ExpiringSet[Sha256Hash]

    def __init__(self, alarm_queue: AlarmQueue, network_num: int = constants.ALL_NETWORK_NUM,) -> None:
        super().__init__(self.NAME, network_num=network_num)

        # enforce uniqueness, since multiple sources can publish to
        # pending transactions (eth ws + remote)
        self.published_transactions = ExpiringSet(
            alarm_queue, EXPIRATION_TIME_S, "pendingTxs"
        )

    def subscribe(self, options: Dict[str, Any]) -> Subscriber[EthTransactionFeedEntry]:
        duplicates = options.get("duplicates", None)
        if duplicates is not None:
            if not isinstance(duplicates, bool):
                raise RpcInvalidParams('"duplicates" must be a boolean')

        return super().subscribe(options)

    def publish(self, raw_message: EthRawTransaction) -> None:
        if (
            raw_message.tx_hash in self.published_transactions
            and not self.any_subscribers_want_duplicates()
        ):
            return

        super().publish(raw_message)

        self.published_transactions.add(raw_message.tx_hash)

    def serialize(self, raw_message: EthRawTransaction) -> EthTransactionFeedEntry:
        return EthTransactionFeedEntry(
            raw_message.tx_hash,
            raw_message.tx_contents,
            raw_message.local_region,
        )

    def any_subscribers_want_duplicates(self) -> bool:
        for subscriber in self.subscribers.values():
            if subscriber.options.get("duplicates", False):
                return True
        return False

    def should_publish_message_to_subscriber(
        self,
        subscriber: Subscriber[EthTransactionFeedEntry],
        raw_message: EthRawTransaction,
        serialized_message: EthTransactionFeedEntry,
    ) -> bool:
        if (
            raw_message.tx_hash in self.published_transactions
            and not subscriber.options.get("duplicates", False)
        ):
            return False
        should_publish = True
        if subscriber.filters:
            logger_filters.trace(
                "checking if should publish to {} with filters {}",
                subscriber.subscription_id,
                subscriber.filters,
            )
            contents = serialized_message.tx_contents
            state = {
                "value": eth_filter_handlers.reformat_tx_value(contents["value"]),
                "to": eth_filter_handlers.reformat_address(contents["to"]),
                "from": eth_filter_handlers.reformat_address(contents["from"]),
                "gas_price": eth_filter_handlers.reformat_gas_price(contents["gas_price"]),
                "max_priority_fee_per_gas": eth_filter_handlers.reformat_gas_price(
                    contents.get("max_priority_fee_per_gas")
                ),
                "max_fee_per_gas": eth_filter_handlers.reformat_gas_price(
                    contents.get("max_fee_per_gas")
                ),
                "method_id": eth_filter_handlers.reformat_input_to_method_id(contents["input"]),
            }
            should_publish = subscriber.validate(state)
            logger_filters.trace("should publish: {}", should_publish)
        return should_publish



from dataclasses import dataclass
from bxgateway.utils.logging.status.analysis import Analysis
from bxgateway.utils.logging.status.summary import Summary


@dataclass
class Diagnostics:
    summary: Summary
    analysis: Analysis



import os
import random
import struct
import sys
from typing import Tuple

import pyelliptic
from Crypto.Cipher import AES
import blxr_rlp as rlp

from bxutils import logging

from bxcommon.exceptions import ParseError

from bxgateway.eth_exceptions import InvalidKeyError, AuthenticationError, CipherNotInitializedError
from bxgateway.messages.eth.serializers.eip8_handshake import Eip8Auth, Eip8AuthAck
from bxcommon.utils.blockchain_utils.eth import crypto_utils, rlp_utils, eth_common_utils, eth_common_constants
from bxgateway.utils.eth.eccx import ECCx
from bxgateway.utils.eth.frame import Frame

logger = logging.get_logger(__name__)


class RLPxCipher(object):

    def __init__(self, is_initiator, private_key, remote_public_key=None):

        self._ecc = ECCx(raw_private_key=private_key)

        if remote_public_key and not self._ecc.is_valid_key(remote_public_key):
            raise InvalidKeyError("Invalid remote pubkey")

        self._ephemeral_ecc = ECCx()

        self._is_initiator = is_initiator
        self._private_key = private_key
        self._remote_pubkey = remote_public_key

        self._is_eip8_auth = True
        self._remote_ephemeral_pubkey = None
        self._initiator_nonce = None
        self._responder_nonce = None
        self._auth_init = None
        self._auth_ack = None
        self._aes_secret = None
        self._token = None
        self._aes_enc = None
        self._aes_dec = None
        self._mac_enc = None
        self._egress_mac = None
        self._ingress_mac = None
        self._is_ready = False
        self._remote_version = 0

    def get_private_key(self):
        """
        Get private key of current ECCx object
        :return: private key
        """

        return self._private_key

    def is_ready(self):
        """
        Returns flag indicating if cipher is ready for encrypted messages exchange
        :return: True if cipher is ready, False otherwise
        """

        return self._is_ready

    def create_auth_message(self):
        """
        Generates authentication message bytes

        1. initiator generates ecdhe-random and nonce and creates auth
        2. initiator connects to remote and sends auth

        New:
        E(remote-pubk,
            S(ephemeral-privk, ecdh-shared-secret ^ nonce) ||
            H(ephemeral-pubk) || pubk || nonce || 0x0
        )
        Known:
        E(remote-pubk,
            S(ephemeral-privk, token ^ nonce) || H(ephemeral-pubk) || pubk || nonce || 0x1)

        :param remote_pubkey: public key of remote node
        :param nonce: nonce
        :return: authentication message bytes
        """

        assert self._is_initiator

        ecdh_shared_secret = self._ecc.get_ecdh_key(self._remote_pubkey)
        token = ecdh_shared_secret
        self._initiator_nonce = eth_common_utils.keccak_hash(rlp_utils.int_to_big_endian(random.randint(0, sys.maxsize)))
        assert len(self._initiator_nonce) == eth_common_constants.SHA3_LEN_BYTES

        token_xor_nonce = crypto_utils.string_xor(token, self._initiator_nonce)
        assert len(token_xor_nonce) == eth_common_constants.SHA3_LEN_BYTES

        ephemeral_pubkey = self._ephemeral_ecc.get_raw_public_key()
        assert len(ephemeral_pubkey) == eth_common_constants.PUBLIC_KEY_LEN
        if not self._ecc.is_valid_key(ephemeral_pubkey):
            raise InvalidKeyError("Invalid ephemeral pubkey")

        # S(ephemeral-privk, ecdh-shared-secret ^ nonce)
        S = self._ephemeral_ecc.sign(token_xor_nonce)
        assert len(S) == eth_common_constants.SIGNATURE_LEN

        version = eth_common_constants.AUTH_MSG_VERSION
        data = rlp.encode(
            (S, crypto_utils.private_to_public_key(self._private_key), self._initiator_nonce, version),
            sedes=Eip8Auth.get_sedes()
        )
        pad = os.urandom(random.randint(eth_common_constants.EIP8_ACK_PAD_MIN, eth_common_constants.EIP8_ACK_PAD_MAX))
        auth_message = data + pad

        return auth_message

    def encrypt_auth_message(self, auth_message):
        """
        Encrypts authentication message
        :param auth_message: message bytes
        :param remote_pubkey: remote public key
        :return: encrypted message
        """

        assert self._is_initiator

        prefix = struct.pack(">H", len(auth_message) + eth_common_constants.ECIES_ENCRYPT_OVERHEAD_LENGTH)
        enc_auth = self._ecc.encrypt(auth_message, self._remote_pubkey, shared_mac_data=prefix)
        self._auth_init = prefix + enc_auth

        return self._auth_init

    def decrypt_auth_message(self, cipher_text):
        """
        Decrypts authentication message
        :param cipher_text: encrypted message
        :return: decrypted authentication message
        """

        assert not self._is_initiator

        if len(cipher_text) < eth_common_constants.ENC_AUTH_MSG_LEN:
            logger.debug("Minimal len of auth message is {0} but was {1}. Continue waiting for more bytes."
                         .format(eth_common_constants.ENC_AUTH_MSG_LEN, len(cipher_text)))
            return None, None

        try:
            if len(cipher_text) == eth_common_constants.ENC_AUTH_MSG_LEN:
                # Decrypt plain auth message
                size = eth_common_constants.ENC_AUTH_MSG_LEN
                message = self._ecc.decrypt(cipher_text[:size])
                self._is_eip8_auth = False
            else:
                # Decrypt EIP8 auth message
                size = struct.unpack(">H", cipher_text[:eth_common_constants.EIP8_AUTH_PREFIX_LEN])[0] + \
                       eth_common_constants.EIP8_AUTH_PREFIX_LEN

                if len(cipher_text) < size:
                    logger.debug("Expected auth message size is {0} but was {1}. Continue waiting more bytes."
                                 .format(size, len(cipher_text)))
                    return None, None

                message = self._ecc.decrypt(bytes(cipher_text[eth_common_constants.EIP8_AUTH_PREFIX_LEN:size]),
                                            shared_mac_data=bytes(cipher_text[:eth_common_constants.EIP8_AUTH_PREFIX_LEN]))
        except RuntimeError as e:
            raise AuthenticationError(e)

        self._auth_init = cipher_text[:size]
        return message, size

    def parse_auth_message(self, message):
        assert not self._is_initiator

        if self._is_eip8_auth:
            (signature, pubkey, nonce, version) = self.parse_eip8_auth_message(message)
        else:
            (signature, pubkey, nonce, version) = self.parse_plain_auth_message(message)

        token = self._ecc.get_ecdh_key(pubkey)
        remote_ephemeral_pubkey = crypto_utils.recover_public_key(crypto_utils.string_xor(token, nonce),
                                                                  signature)
        if not self._ecc.is_valid_key(remote_ephemeral_pubkey):
            raise InvalidKeyError("Invalid remote ephemeral pubkey")

        self._remote_ephemeral_pubkey = remote_ephemeral_pubkey
        self._initiator_nonce = nonce
        self._remote_pubkey = pubkey
        self._remote_version = eth_common_constants.P2P_PROTOCOL_VERSION

    def parse_plain_auth_message(self, message):
        if len(message) != eth_common_constants.AUTH_MSG_LEN:
            raise ValueError("Authentication message of len {0} is expected but was {1}"
                             .format(eth_common_constants.ENC_AUTH_MSG_LEN, len(message)))

        signature = message[:eth_common_constants.SIGNATURE_LEN]

        pubkey_start = eth_common_constants.SIGNATURE_LEN + eth_common_constants.MAC_LEN
        pubkey = message[pubkey_start:pubkey_start + eth_common_constants.PUBLIC_KEY_LEN]
        if not self._ecc.is_valid_key(pubkey):
            raise InvalidKeyError("Invalid initiator pubkey")

        nonce_start = eth_common_constants.SIGNATURE_LEN + eth_common_constants.MAC_LEN + eth_common_constants.PUBLIC_KEY_LEN
        nonce = message[nonce_start: nonce_start + eth_common_constants.AUTH_NONCE_LEN]
        known_flag = bool(rlp_utils.safe_ord(message[nonce_start + eth_common_constants.AUTH_NONCE_LEN:]))
        assert known_flag == 0

        return signature, pubkey, nonce, eth_common_constants.P2P_PROTOCOL_VERSION

    def parse_eip8_auth_message(self, message):

        values = rlp.decode(message, sedes=Eip8Auth.get_sedes(), strict=False)
        signature, pubkey, nonce, version = values

        return signature, pubkey, nonce, version

    def create_auth_ack_message(self):
        """
        authRecipient = E(remote-pubk, remote-ephemeral-pubk || nonce || 0x1) // token found
        authRecipient = E(remote-pubk, remote-ephemeral-pubk || nonce || 0x0) // token not found

        nonce, ephemeral_pubkey, version are local!
        """

        assert not self._is_initiator

        ephemeral_pubkey = self._ephemeral_ecc.get_raw_public_key()
        encoded_nonce = rlp.sedes.big_endian_int.serialize(random.randint(0, eth_common_constants.MAX_NONCE))
        self._responder_nonce = eth_common_utils.keccak_hash(encoded_nonce)

        if self._is_eip8_auth:
            msg = self.create_eip8_auth_ack_message(ephemeral_pubkey, self._responder_nonce,
                                                    eth_common_constants.P2P_PROTOCOL_VERSION)
            assert len(msg) > eth_common_constants.AUTH_ACK_MSG_LEN
        else:
            msg = self.create_plain_auth_ack_message(ephemeral_pubkey, self._responder_nonce,
                                                     eth_common_constants.P2P_PROTOCOL_VERSION)
            assert len(msg) == eth_common_constants.AUTH_ACK_MSG_LEN

        return msg

    def create_plain_auth_ack_message(self, ephemeral_pubkey, nonce, version):
        return ephemeral_pubkey + nonce + b"\x00"

    def create_eip8_auth_ack_message(self, ephemeral_pubkey, nonce, version):

        data = rlp.encode((ephemeral_pubkey, nonce, version), sedes=Eip8AuthAck.get_sedes())
        pad = os.urandom(random.randint(eth_common_constants.EIP8_ACK_PAD_MIN, eth_common_constants.EIP8_ACK_PAD_MAX))
        return data + pad

    def encrypt_auth_ack_message(self, ack_message):
        assert not self._is_initiator

        if self._is_eip8_auth:
            prefix = struct.pack(">H", len(ack_message) + eth_common_constants.ECIES_ENCRYPT_OVERHEAD_LENGTH)
            enc_auth_ack = self._ecc.encrypt(ack_message, self._remote_pubkey, shared_mac_data=prefix)

            self._auth_ack = prefix + enc_auth_ack
        else:
            self._auth_ack = self._ecc.encrypt(ack_message, self._remote_pubkey)
            assert len(self._auth_ack) == eth_common_constants.ENC_AUTH_ACK_MSG_LEN

        return self._auth_ack

    def decrypt_auth_ack_message(self, cipher_text):
        """
        Decrypts authentication acknowledge message
        :param cipher_text: encrypted message
        :return: decrypted message
        """

        assert self._is_initiator

        if not self._is_eip8_auth and len(cipher_text) != eth_common_constants.ENC_AUTH_ACK_MSG_LEN:
            raise ValueError("Expected length of auth ack message is {0} but was provided {1}"
                             .format(eth_common_constants.ENC_AUTH_ACK_MSG_LEN, len(cipher_text)))

        if self._is_eip8_auth:
            (ephemeral_pubkey, nonce, version) = self._decode_eip8_auth_ack_message(cipher_text)
            self._auth_ack = cipher_text
            self._remote_ephemeral_pubkey = ephemeral_pubkey
            self._responder_nonce = nonce
            self._remote_version = version
            return cipher_text
        else:
            (size, eph_pubkey, nonce) = self._decode_auth_ack_message(cipher_text)
            self._auth_ack = cipher_text[:size]
            self._remote_ephemeral_pubkey = eph_pubkey[:eth_common_constants.PUBLIC_KEY_LEN]
            self._responder_nonce = nonce
            self._remote_version = eth_common_constants.AUTH_MSG_VERSION

            if not self._ecc.is_valid_key(self._remote_ephemeral_pubkey):
                raise InvalidKeyError("Invalid remote ephemeral pubkey")

            return cipher_text[size:]

    def setup_cipher(self):
        """
        Sets up cipher parameters.
        Needs to be called after initial authentication handshake
        """

        assert self._responder_nonce
        assert self._initiator_nonce
        assert self._auth_init
        assert self._auth_ack
        assert self._remote_ephemeral_pubkey
        if not self._ecc.is_valid_key(self._remote_ephemeral_pubkey):
            raise InvalidKeyError("Invalid remote ephemeral pubkey")

        # derive base secrets from ephemeral key agreement
        # ecdhe-shared-secret = ecdh.agree(ephemeral-privkey, remote-ephemeral-pubk)
        ecdhe_shared_secret = self._ephemeral_ecc.get_ecdh_key(self._remote_ephemeral_pubkey)

        # shared-secret = sha3(ecdhe-shared-secret || sha3(nonce || initiator-nonce))
        shared_secret = eth_common_utils.keccak_hash(
            ecdhe_shared_secret + eth_common_utils.keccak_hash(self._responder_nonce + self._initiator_nonce))

        self.ecdhe_shared_secret = ecdhe_shared_secret  # used in tests
        self.shared_secret = shared_secret  # used in tests

        # token = sha3(shared-secret)
        self._token = eth_common_utils.keccak_hash(shared_secret)

        # aes-secret = sha3(ecdhe-shared-secret || shared-secret)
        self._aes_secret = eth_common_utils.keccak_hash(ecdhe_shared_secret + shared_secret)

        # mac-secret = sha3(ecdhe-shared-secret || aes-secret)
        self.mac_secret = eth_common_utils.keccak_hash(ecdhe_shared_secret + self._aes_secret)

        # setup sha3 instances for the MACs
        # egress-mac = sha3.update(mac-secret ^ recipient-nonce || auth-sent-init)
        mac1 = crypto_utils.get_sha3_calculator(
            crypto_utils.string_xor(self.mac_secret, self._responder_nonce) + self._auth_init)
        # ingress-mac = sha3.update(mac-secret ^ initiator-nonce || auth-recvd-ack)
        mac2 = crypto_utils.get_sha3_calculator(
            crypto_utils.string_xor(self.mac_secret, self._initiator_nonce) + self._auth_ack)

        if self._is_initiator:
            self._egress_mac, self._ingress_mac = mac1, mac2
        else:
            self._egress_mac, self._ingress_mac = mac2, mac1

        iv = "\x00" * eth_common_constants.IV_LEN
        self._aes_enc = pyelliptic.Cipher(self._aes_secret, iv, eth_common_constants.CIPHER_ENCRYPT_DO,
                                          ciphername=eth_common_constants.RLPX_CIPHER_NAME)
        self._aes_dec = pyelliptic.Cipher(self._aes_secret, iv, eth_common_constants.CIPHER_DECRYPT_DO,
                                          ciphername=eth_common_constants.RLPX_CIPHER_NAME)
        self._mac_enc = AES.new(self.mac_secret, AES.MODE_ECB).encrypt

        self._is_ready = True

    def encrypt_frame(self, frame):
        """
        Encrypts frame
        :param frame: frame data
        :return: encrypted frame
        """

        if not isinstance(frame, Frame):
            raise TypeError("frame must be of type Frame but was {0}".format(type(frame)))

        if not self._is_ready:
            raise CipherNotInitializedError(f"failed to encrypt frame {frame}, the cipher was never initialized!")

        header = frame.get_header()
        body = frame.get_body()

        # header
        header_ciphertext = self.aes_encode(header)

        # egress-mac.update(aes(mac-secret,egress-mac) ^ header-ciphertext).digest
        header_mac = self.mac_egress(
            crypto_utils.string_xor(self._mac_enc(self.mac_egress()[:eth_common_constants.FRAME_MAC_LEN]),
                                    header_ciphertext))[:eth_common_constants.FRAME_MAC_LEN]

        # frame
        frame_ciphertext = self.aes_encode(body)
        assert len(frame_ciphertext) == len(body)
        # egress-mac.update(aes(mac-secret,egress-mac) ^
        # left128(egress-mac.update(frame-ciphertext).digest))
        fmac_seed = self.mac_egress(frame_ciphertext)
        frame_mac = self.mac_egress(
            crypto_utils.string_xor(self._mac_enc(self.mac_egress()[:eth_common_constants.FRAME_MAC_LEN]),
                                    fmac_seed[:eth_common_constants.FRAME_MAC_LEN]))[:eth_common_constants.FRAME_MAC_LEN]

        return bytearray(header_ciphertext + header_mac + frame_ciphertext + frame_mac)

    def decrypt_frame_header(self, data):
        """
        Decrypts frame header
        :param data: frame header data
        :return: decrypted header data
        """

        if len(data) != eth_common_constants.FRAME_HDR_TOTAL_LEN:
            raise ValueError("Frame header len is expected to be {0} but was {1}"
                             .format(eth_common_constants.FRAME_HDR_TOTAL_LEN, len(data)))

        if not self._is_ready:
            raise CipherNotInitializedError("failed to decrypt frame header, the cipher was never initialized!")

        header_cipher_text = data[:eth_common_constants.FRAME_HDR_DATA_LEN]
        header_mac = data[eth_common_constants.FRAME_HDR_DATA_LEN:eth_common_constants.FRAME_HDR_TOTAL_LEN]

        # ingress-mac.update(aes(mac-secret,ingress-mac) ^ header-ciphertext).digest
        mac1 = self.mac_ingress()[:eth_common_constants.FRAME_MAC_LEN]
        mac_enc = self._mac_enc(mac1)
        mac_sxor = crypto_utils.string_xor(mac_enc, header_cipher_text)

        expected_header_mac = self.mac_ingress(mac_sxor)[:eth_common_constants.FRAME_MAC_LEN]

        # expected_header_mac = self.updateMAC(self.ingress_mac, header_ciphertext)
        if not expected_header_mac == header_mac:
            raise AuthenticationError("Invalid header mac")

        return self.aes_decode(header_cipher_text)

    def decrypt_frame_body(self, data, body_size):
        """
        Decrypts frame body
        :param data: frame data
        :param body_size: body size
        :return: decrypted frame body
        """

        if not self._is_ready:
            raise CipherNotInitializedError("failed to decrypt frame body, the cipher was never initialized!")


        # frame-size: 3-byte integer size of frame, big endian encoded (excludes padding)
        # frame relates to body w/o padding w/o mac

        read_size = crypto_utils.get_padded_len_16(body_size)

        if not len(data) >= read_size + eth_common_constants.FRAME_MAC_LEN:
            raise ParseError("Insufficient body length")

        frame_cipher_text = data[:read_size]
        frame_mac = data[read_size:read_size + eth_common_constants.FRAME_MAC_LEN]

        # ingres-mac.update(aes(mac-secret,ingres-mac) ^
        # left128(ingres-mac.update(frame-ciphertext).digest))
        frame_mac_seed = self.mac_ingress(frame_cipher_text)
        expected_frame_mac = self.mac_ingress(
            crypto_utils.string_xor(self._mac_enc(self.mac_ingress()[:eth_common_constants.FRAME_MAC_LEN]),
                                    frame_mac_seed[:eth_common_constants.FRAME_MAC_LEN]))[:eth_common_constants.FRAME_MAC_LEN]

        if not frame_mac == expected_frame_mac:
            raise AuthenticationError("Invalid frame mac")

        return self.aes_decode(frame_cipher_text)[:body_size]

    def aes_encode(self, data=""):
        if isinstance(data, bytearray):
            bytes_to_encode = bytes(data)
        elif isinstance(data, memoryview):
            bytes_to_encode = data.tobytes()
        else:
            bytes_to_encode = data

        return self._aes_enc.update(bytes_to_encode)

    def aes_decode(self, data=""):
        if isinstance(data, bytearray):
            bytes_to_decode = bytes(data)
        elif isinstance(data, memoryview):
            bytes_to_decode = data.tobytes()
        else:
            bytes_to_decode = data

        return self._aes_dec.update(bytes_to_decode)

    def mac_egress(self, data=b""):
        data = rlp_utils.str_to_bytes(data)
        self._egress_mac.update(data)
        return self._egress_mac.digest()

    def mac_ingress(self, data=b""):
        data = rlp_utils.str_to_bytes(data)
        self._ingress_mac.update(data)
        return self._ingress_mac.digest()

    def _decode_auth_ack_message(self, cipher_text):
        try:
            message = self._ecc.decrypt(cipher_text[:eth_common_constants.ENC_AUTH_ACK_MSG_LEN])
        except RuntimeError as e:
            raise AuthenticationError(e)
        eph_pubkey = message[:eth_common_constants.PUBLIC_KEY_LEN]
        nonce = message[eth_common_constants.PUBLIC_KEY_LEN:eth_common_constants.PUBLIC_KEY_LEN + eth_common_constants.AUTH_NONCE_LEN]
        known = rlp_utils.safe_ord(message[-1])
        assert known == 0
        return (eth_common_constants.ENC_AUTH_ACK_MSG_LEN, eph_pubkey, nonce)

    def _decode_eip8_auth_ack_message(self, cipher_text: bytes) -> Tuple[bytes, bytes, int]:
        message = self._ecc.decrypt(bytes(cipher_text[eth_common_constants.EIP8_AUTH_PREFIX_LEN:]),
                                    shared_mac_data=bytes(cipher_text[:eth_common_constants.EIP8_AUTH_PREFIX_LEN]))

        values = rlp.decode(message, sedes=Eip8AuthAck.get_sedes(), strict=False)
        (ephemeral_pubkey, nonce, version) = values
        if not self._ecc.is_valid_key(ephemeral_pubkey):
            raise InvalidKeyError("Invalid remote ephemeral pubkey")
        return ephemeral_pubkey, nonce, version
